import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.PrintStream;
import java.io.PrintWriter;
import java.io.StringWriter;
import java.nio.charset.StandardCharsets;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Scanner;
import java.util.StringTokenizer;

import org.apache.commons.math3.distribution.*;

public class Experiment {
	//private static HashMap<Integer, ArrayList<Integer>> faultMatrix;
	private  HashMap<Integer, ArrayList<Integer>> previousFaultMatrixRT;
	private  HashMap<Integer, ArrayList<Integer>> previousFaultMatrixOT;
	private  HashMap<Integer,Double>  occurrenceProbability;
	private  ArrayList<Integer> failedTestCases;
	private  ArrayList<Integer> failedTestCases_OT;
	private  ArrayList<Integer> potentialTestCases_OT;
	private  ArrayList<Integer> selectedTestCasesRT;
	private  ArrayList<Integer> cumulativeSelectedTestCasesRT;
	private  ArrayList<Integer> cumulativeFailedTestCases_RT;
	private  ArrayList<Integer> cumulativeSelectedTestCasesOT;
	private  ArrayList<Integer> cumulativeFailedTestCases_OT;
	private  ArrayList<Integer> cumulativeNumberFaultsOP;
	private  ArrayList<Integer> cumulativeNumberFaultsRT;
	private  int numberDetectedFaults;
	private ArrayList<Double> reliabiltiyGrowthList;
	private ArrayList<Double> reliabiltiyGrowthList_RT;
	private ArrayList<Double> reliabiltiyGrowthList_OT;
	ArrayList<Integer> progressiveTestAtRelGrowth_OT; 
	ArrayList<Integer> progressiveTestAtRelGrowth_RT; 
	

	//IS ALGORITHM PARAMTERS
	double significance;
	double desiredError;
	double learningFactor;
	
	/*********PROGRAM - DEPENDENT PARAMTERS: EXPERIMENT INPUT *********/
	int totalTests;  
	int numberOfPartitions;
	int[] domainSize; 	
	double[] p_i; 
	String program_name;
	String version_name;
	
	/*********EXPERIMENT INPUT *********/
	
	String coverageCriterion;
	String faultMatrix;
	
	/*********EXPERIMENT OUPUT *********/

	double reliability;	
	double reliabilityOT;
	ArrayList<Double> reliabilitiesSequence_OT;  
	ArrayList<Double> reliabilitiesSequence_RT;
	ArrayList<Integer> testCaseSequence;  
	private  FileWriter output;
	private  FileWriter csv_output;
	private  FileWriter csv_2_output;
	private  FileWriter csv_4_output;
	private  FileWriter csv_5_output;
	
	private  PrintWriter pw;
	private  PrintWriter csv_w;
	private  PrintWriter csv_w_2;
	private  PrintWriter csv_w_4;
	private  PrintWriter csv_w_5;
	String csv_string;
		
	public Experiment(int _totalTests, int _numberOfPartitions, String _program_name, String _version_name, int[] _domainSize, String outputFile, double _significance, double _desiredError, double _learningFactor, String _coverageCriterion, String _faultMatrix){
	
		/*********PROGRAM - DEPENDENT PARAMTERS *********/
		totalTests=  _totalTests;
		numberOfPartitions = _numberOfPartitions;
	
		program_name = _program_name;
		version_name = _version_name;
		domainSize = new int[numberOfPartitions]; 	
		p_i = new double[numberOfPartitions]; 
		
		for(int i=0;i<numberOfPartitions;i++){
			p_i[i] = -1; // to be set up by SET PROFILE
			domainSize[i] = _domainSize[i];
		}

		significance = _significance;
		desiredError = _desiredError;
		learningFactor = _learningFactor;
		coverageCriterion = _coverageCriterion;
		faultMatrix = _faultMatrix;
		
		/* EXPERIMENT OUTPUT */
		reliability = 0;	
		reliabilityOT = 0;
		reliabilitiesSequence_OT = new ArrayList<Double>();  
		reliabilitiesSequence_RT = new ArrayList<Double>();
		testCaseSequence = new ArrayList<Integer>();  
		
		String csv_outputFile = outputFile.replace("txt", "csv"); 
		csv_outputFile = csv_outputFile.replaceAll("output", "csv_Iterations");
		System.out.println("csv_outputFile "+csv_outputFile);
		
		String csv_2_outputFile = outputFile.replace("txt", "csv"); 
		csv_2_outputFile = csv_2_outputFile.replace("output", "csv_Reps");
		System.out.println("csv_2_outputFile "+csv_2_outputFile);
		
		//CSV 3 is created  in the MAIN
		
		String csv_4_outputFile = outputFile.replace("txt", "csv"); 
		csv_4_outputFile = csv_4_outputFile.replace("output", "csv_RelGrowthAtFailure");
		System.out.println("csv_4_outputFile "+csv_4_outputFile);
		
		String csv_5_outputFile = outputFile.replace("txt", "csv"); 
		csv_5_outputFile = csv_5_outputFile.replace("output", "csv_NTestsBeforeFailure");
		System.out.println("csv_5_outputFile "+csv_5_outputFile);
		
		try {
			output  = new FileWriter(outputFile, true);
			csv_output  = new FileWriter(csv_outputFile, true);
			csv_2_output  = new FileWriter(csv_2_outputFile, true);
			csv_4_output  = new FileWriter(csv_4_outputFile, true);
			csv_5_output  = new FileWriter(csv_5_outputFile, true);
		} catch (IOException e) {e.printStackTrace();}
		pw = new PrintWriter(new BufferedWriter(output));
		csv_w = new PrintWriter(new BufferedWriter(csv_output));
		csv_w_2 = new PrintWriter(new BufferedWriter(csv_2_output));
		csv_w_4 = new PrintWriter(new BufferedWriter(csv_4_output));
		csv_w_5 = new PrintWriter(new BufferedWriter(csv_5_output));
		
		this.cumulativeFailedTestCases_OT 	= new ArrayList<Integer>();
		this.cumulativeSelectedTestCasesOT 	= new ArrayList<Integer>();
		this.cumulativeFailedTestCases_RT 	= new ArrayList<Integer>();
		this.cumulativeSelectedTestCasesRT 	= new ArrayList<Integer>();
		this.cumulativeNumberFaultsOP 		= new ArrayList<Integer>();
		this.cumulativeNumberFaultsRT 		= new ArrayList<Integer>();
		this.numberDetectedFaults			= 0;
		this.reliabiltiyGrowthList  		= new ArrayList<Double>();
		this.reliabiltiyGrowthList_RT  		= new ArrayList<Double>();
		this.reliabiltiyGrowthList_OT  		= new ArrayList<Double>();
		this.progressiveTestAtRelGrowth_OT	 =new ArrayList<Integer>(); 
		this.progressiveTestAtRelGrowth_RT 	=new ArrayList<Integer>(); 
		
	};
	
	public void setProfile(double[] probabilities){
		for(int i=0; i<this.numberOfPartitions;i++)
			p_i[i] = probabilities[i];
	}
	
	public int runSession() throws IOException, InterruptedException{
		
		// VARIABLES USED DURING THE SESSION BY IS
		double[] probabilities 			= new double[numberOfPartitions]; // represent the probability used to assign test cases to domains (it is the pi_greco value of Equation 15 of my TSE paper)
		double[] oldProbabilities 		= new double[numberOfPartitions];
 		double[] numberOfFailures 		= new double[numberOfPartitions]; 
		double[] failureRate 			= new double[numberOfPartitions];
		double[] unrelContribution 		= new double[numberOfPartitions];
		double[] TestCasePerPartition 	= new double[numberOfPartitions];
		double[] sumExecutedTestCasePerPartition = new double[numberOfPartitions];
		double[] numberOfPotentialTests 	= new double[numberOfPartitions];
		boolean withoutReplacement = true;
		boolean winningSet = false;
		boolean achievedMaxReliabilityRT = false;
		boolean achievedMaxReliabilityOT = false;
		boolean[] partitionFull = new boolean[numberOfPartitions];
		for(int i=0; i < numberOfPartitions; i++)
			 partitionFull[i]=false;
			
		if (p_i[0]== -1){
			System.out.println("ERROR: the SET_PROFILE function should be called first, in order to have a profile defined ");
			System.exit(0);
		}
		this.cumulativeFailedTestCases_OT.clear();
		this.cumulativeSelectedTestCasesOT.clear();
		this.cumulativeFailedTestCases_RT.clear(); 
		this.cumulativeSelectedTestCasesRT.clear(); 
		this.cumulativeNumberFaultsOP.clear();
		this.cumulativeNumberFaultsRT.clear();
		this.testCaseSequence.clear();
		this.reliabilitiesSequence_OT.clear();
		this.reliabilitiesSequence_RT.clear();
		this.numberDetectedFaults			= 0;
		this.reliabiltiyGrowthList.clear();
		this.reliabiltiyGrowthList_RT.clear();
		this.reliabiltiyGrowthList_OT.clear();
		this.progressiveTestAtRelGrowth_OT.clear(); 
		this.progressiveTestAtRelGrowth_RT.clear();
		
		Path filePath = Paths.get(faultMatrix); 	
		//PRINT SOME INFO
		System.out.format("Selected Fault Matrx path: %s%n", filePath.toString());
		System.out.format("Selected Fault Matrix name: %s%n", filePath.getFileName());
		//System.out.format("getParent: %s%n", filePath.getParent());
		//System.out.format("getRoot: %s%n", filePath.getRoot());
	
		
		int indexCurrentTest = 0;	
		previousFaultMatrixRT = loadFaultMatrix(filePath);
		previousFaultMatrixOT = loadFaultMatrix(filePath);
		occurrenceProbability = computeOccurrenceProbability(domainSize,totalTests, p_i);
		failedTestCases = new ArrayList<Integer>();
		potentialTestCases_OT = new ArrayList<Integer>();
		for(int i=1;i<=this.totalTests;i++) //INITIALIZE: ASSUME THAT TESTS ARE IDENTIFIED BY PROGRESSIVE INTEGER
			potentialTestCases_OT.add(i);
			
		for (int i=0; i<numberOfPartitions;i++ ){
			numberOfPotentialTests[i] = domainSize[i];
			oldProbabilities[i] = 0;
			probabilities[i] = p_i[i];
			numberOfFailures[i] = 0;
			failureRate[i]=0;
			unrelContribution[i] = 0;
			TestCasePerPartition[i] = 0;
			sumExecutedTestCasePerPartition[i] = 0;			
		}
	
	/********************** ITERATION 0 - WITHOUT IMPORTANCE SAMPLING ALGO**********************/
		
	int initialNumberOfTestCases=0; 
	double sumTC_overPartitions = 0;
	boolean allZeroFailure = true;
	boolean retried = false;
	int[] countIteration = new int[numberOfPartitions];
	for (int k=0; k<numberOfPartitions;k++)
		countIteration[k] = 0;
	
	int returnValue=-2;//return error by default - unsuccessful execution

//	while(allZeroFailure){ 	 // IF WE HAVE ZERO FAILURES IN ALL THE PARTITIONS, THEN RE-USE THE SAME ALLOCATION
		
		initialNumberOfTestCases = numberOfPartitions; //we can change This	    
		sumTC_overPartitions = 0;
	
		 /* OPERATIONAL TESTING WITH THE SAME NUMBER OF TEST CASES*/
		
		// Number of Tests PROPORTIONALLY TO DOMAIN SIZE (BUT I DON'T USE THESE TESTS YET)
		//ONLY TO DETERMINE HOW MANY TESTS PER PARTITION I HAVE TO TAKE IN RT
		withoutReplacement = false; // THESE ARE NOT EXECUTED, SO THEY MUST NOT BE REMOVED (with replacement)
		ArrayList<Integer> selecetdTestCasesPerPartition = selectTestCases_OT(initialNumberOfTestCases, withoutReplacement);
		// DO NOT UPDATE potentialTestCases_OT, BECAUSE THESE TESTS are not executed actually
		// DETERMINE WHICH PARTITION THEY BELONG TO
		 double upperLimit;
		 for (int k=0; k<numberOfPartitions;k++) TestCasePerPartition[k]=0;
		 for (int j=0; j<selecetdTestCasesPerPartition.size();j++){
			 upperLimit=domainSize[0];
			  for (int k=0; k<numberOfPartitions;k++){
				 if(selecetdTestCasesPerPartition.get(j) <= upperLimit){
					 TestCasePerPartition[k]++;
					 break;
				 }
				 else
					 upperLimit = upperLimit + domainSize[k+1]; 	
			 	}
			 }
		 // SELECT AND RUN TEST CASES FROM RT 
		 System.out.println("Working Directory = " + System.getProperty("user.dir"));
			//if (retried==false)
				Runtime.getRuntime().exec("python select.py "+program_name+" --ver "+version_name+" --cleanup");
			 selectedTestCasesRT = new ArrayList<Integer>();//clear();
			 failedTestCases.clear();
			 String command= null; 
			 double sumOverPartitions = 0;
			 for (int k=0; k<numberOfPartitions;k++){
				 if (TestCasePerPartition[k]!=0){
				 System.out.println("Initial Number of Test cases per partition "+k+" :"+TestCasePerPartition[k]);
				 command="python select.py "+program_name+" --ver "+version_name+" -c "+coverageCriterion+" --fm all -i "+countIteration[k]+" -p "+k+" -n "+(int)TestCasePerPartition[k]+"";
				 countIteration[k]++;
				 System.out.println("Command "+command);
				 Process p = Runtime.getRuntime().exec(command);
				 BufferedReader br = new BufferedReader(new InputStreamReader(p.getInputStream()));
				 BufferedReader stdError = new BufferedReader(new InputStreamReader(p.getErrorStream()));
				 int returnCode = p.waitFor();
				if (returnCode >= 2) {
					System.out.println("OS Error: Unable to Find File or other OS error.");
				}   	
				String str =null;  
				while (br.ready()) {
					str = br.readLine();
					System.out.println("Reading...");
					System.out.println(str);
				}
				StringTokenizer st = new StringTokenizer(str, ", ");
				while(st.hasMoreTokens())
					selectedTestCasesRT.add(Integer.parseInt(st.nextToken().trim()));			
				
				System.out.println("Selected Test Cases by RT "+selectedTestCasesRT);
				failedTestCases = getFailedTests(previousFaultMatrixRT, selectedTestCasesRT);
				System.out.println("FAILED TEST CASES IN ADAPTIVE RELIABILITY TESTING: "+failedTestCases);		
				
				 if(k==0){
					 numberOfFailures[k] = failedTestCases.size();
					 sumOverPartitions = numberOfFailures[k];
				 }
				 else{
					 numberOfFailures[k] = failedTestCases.size() - sumOverPartitions;	
					 sumOverPartitions = failedTestCases.size();
					 }
		 		}
				 else{
						numberOfFailures[k]=0;
					}
					
			 }
			 this.cumulativeFailedTestCases_RT.addAll(failedTestCases);
				
			 this.cumulativeSelectedTestCasesRT.addAll(selectedTestCasesRT);
				
		 //SEELCT ACTUAL TEST CASES FOR OT
		if (!previousFaultMatrixOT.isEmpty()){
			failedTestCases_OT = new ArrayList<Integer>();
			int numberOfTestCases_OT = initialNumberOfTestCases; 
			ArrayList<Integer> selecetdTestCasesOT = new ArrayList<Integer>(selectedTestCasesRT);// selectTestCases_OT(numberOfTestCases_OT, occurrenceProbability);
			System.out.println("SELECTED TEST CASES BY OPERATIONAL TESTING: "+selecetdTestCasesOT);
			//REMOVE FROM POTENTIAL LIST
			for(int i=0; i< selecetdTestCasesOT.size();i++) potentialTestCases_OT.remove(selecetdTestCasesOT.get(i));
			this.cumulativeSelectedTestCasesOT.addAll(selecetdTestCasesOT);
			failedTestCases_OT = getFailedTests(previousFaultMatrixOT, selecetdTestCasesOT);
			System.out.println("FAILED TEST CASES IN OPERATIONAL TESTING: "+failedTestCases_OT);
			this.cumulativeFailedTestCases_OT.addAll(failedTestCases_OT);
			this.numberDetectedFaults = 0;
			this.reliabiltiyGrowthList.clear();
			HashMap<Integer, ArrayList<Integer>> updatedFaultMatrixOT = updateFaultMatrix(failedTestCases_OT,previousFaultMatrixOT);
			previousFaultMatrixOT = updatedFaultMatrixOT; 
			this.reliabiltiyGrowthList_OT.addAll(this.reliabiltiyGrowthList);
			this.cumulativeNumberFaultsOP.add(this.numberDetectedFaults);
			System.out.println("cumulativeNumberFaultsOP "+cumulativeNumberFaultsOP);

			reliabilityOT = computeDeliveredReliability(updatedFaultMatrixOT);
			System.out.println("\nReliability of Operational Testing after "+numberOfTestCases_OT+" test cases(Iteration 0): "+reliabilityOT+"\n");
		}
	
		/* COMPUTE DELIVERED RELIABILITY AT ITERATION 0 */
			
		this.numberDetectedFaults = 0;
		this.reliabiltiyGrowthList.clear();
		 HashMap<Integer, ArrayList<Integer>> updatedFaultMatrixRT = updateFaultMatrix(failedTestCases,previousFaultMatrixRT);
		 System.out.println("FAILED TEST CASES IN ADAPTIVE RELIABILITY TESTING: "+failedTestCases);
		 previousFaultMatrixRT = updatedFaultMatrixRT; 
		 this.reliabiltiyGrowthList_RT.addAll(this.reliabiltiyGrowthList);
		 this.cumulativeNumberFaultsRT.add(this.numberDetectedFaults);
		 System.out.println("cumulativeNumberFaultsRT "+cumulativeNumberFaultsRT);
		 reliability = computeDeliveredReliability(updatedFaultMatrixRT);
		 System.out.println("\nReliability of Adaptive Reliability Testing at Iteration 0: "+reliability+"\n");
		/* COMPUTE DELIVERED RELIABILITY AT ITERATION 0  - END*/
		 
		 
		  for (int i=0; i<numberOfPartitions;i++ ) {
			 indexCurrentTest = (int) (indexCurrentTest + TestCasePerPartition[i]); //EXECUTED TESTS
			 sumExecutedTestCasePerPartition[i] += TestCasePerPartition[i]; 
		 
		 }
		 System.out.println("Test Cases for Iteration 0: "+initialNumberOfTestCases+" \n\n");
		 for (int i=0;i<numberOfPartitions;i++)System.out.println("TestCasePerPartition "+i+" at iteration 0:  "+TestCasePerPartition[i]+"\n");
		 for (int i=0;i<numberOfPartitions;i++)System.out.println("\nnumberOfFailures in partition "+i+" at iteration 0:  "+numberOfFailures[i]);

		 reliabilitiesSequence_OT.add(reliabilityOT);
		 reliabilitiesSequence_RT.add(reliability);
		 testCaseSequence.add(indexCurrentTest);
		 
		 
		 int checkZeroFailure=0;
		 for (int i=0; i<numberOfPartitions;i++ ) {
			 checkZeroFailure += numberOfFailures[i];
			 if (checkZeroFailure != 0)
					 allZeroFailure = false; //IF THERE HAS BEEN NO FAILURE, RETRY WITH THE SAME ALLOCATION
			 else 
				 retried=true;
		 }
		 System.out.println("indexCurrentTest: "+indexCurrentTest);
		 System.out.println("END OF ITERATION 0 \n\n"); 
		 if (indexCurrentTest >= totalTests){
			 System.out.println("indexCurrentTest: "+indexCurrentTest);
			 System.out.println("ERROR: STOP THE TEST WITHOUT ANY CALL TO IMPORTANCE SAMPLING. The approach in this case is useless \n\n");
			 System.exit(0); 
		 }
	
	
	//	 for (int i=0;i<numberOfPartitions;i++)
	//			 numberOfFailures[i]=0;
	//}	 
	 
		 
		 
		
		 if(reliability==1 && reliabilityOT==1){
			 System.out.println("\n\nTEST ENDED AT ITERATION 0");
			 computeInterFailureTests();
			 printResults();
			 if(this.progressiveTestAtRelGrowth_RT.get(this.progressiveTestAtRelGrowth_RT.size()-1) < this.progressiveTestAtRelGrowth_OT.get(this.progressiveTestAtRelGrowth_OT.size()-1))
				 return 1;
			 if(this.progressiveTestAtRelGrowth_RT.get(this.progressiveTestAtRelGrowth_RT.size()-1) > this.progressiveTestAtRelGrowth_OT.get(this.progressiveTestAtRelGrowth_OT.size()-1))
				 return 0;
			 if(this.progressiveTestAtRelGrowth_RT.get(this.progressiveTestAtRelGrowth_RT.size()-1) == this.progressiveTestAtRelGrowth_OT.get(this.progressiveTestAtRelGrowth_OT.size()-1))
				 return -1;
			
		 }
		 if(winningSet == false){
			 if (reliability==1 && reliabilityOT<1){
				 //System.out.println("\n\nTEST ENDED AT ITERATION 0");
				// printResults();
				// return 1;
				 winningSet = true;
				 returnValue = 1;
				 achievedMaxReliabilityRT = true;
			 }
			 if (reliabilityOT==1 && reliability<1){
				 //System.out.println("\n\nTEST ENDED AT ITERATION 0");
				 //printResults();
				 //return 0;
				 winningSet = true;
				 returnValue = 0;
				 achievedMaxReliabilityOT = true;
			}
		}
	
		 
		 
		/********************** AT NEXT ITERATIONS- algorithm IS **********************/   
	
	/*** THE ALGORITHM TENDS TO ASSIGN MORE TESTS TO PARTITIONS THAT ARE EXPERIENCING HIGHER FAILURE RATE (THUS IT PRESUMES THAT IN THE NEXT ITERATION THE FAILURE RATE WILL BE STILL HIGH. WHEN IT STARTS DECREASING, THE SUCCESSIVE ITERATIONS WILL ASSIGN (RELATIVELY) LESS TEST CASES TO IT)  ****/

		 int Iteration = 1;
		 boolean achievedMaxReliability = false;
		 while(indexCurrentTest < totalTests && achievedMaxReliability==false) {

			 int numberOfTestCasesNextIteration = 0;
			 
if(achievedMaxReliabilityRT==false){		 
			 System.out.println("********* COMPUTING DATA FOR ITERATION "+Iteration+"*******\n"); 
			 double sumUnrelContribution = 0;
			 double sumFailureRate = 0;
			 for (int i=0; i<numberOfPartitions;i++ ){
				 if (TestCasePerPartition[i]==0)
						 failureRate[i]=0;
				 else
					 failureRate[i] =  (numberOfFailures[i])/TestCasePerPartition[i];  // Failure rate estimation for subdomain  "i". Assume that all the tests suggested per partition have been executed
				 // NORAMZLIZeD FAILURE RATE
				 
				 sumFailureRate +=failureRate[i];
				 //unrelContribution[i] = failureRate[i]*p_i[i];    // this is p_i * FailureRate_i  ... it represents the expected contribution to unreliabiltiy expected if we select a test case from subdomain i. The assignment of test cases is given proportionally to this value.
				 //sumUnrelContribution += unrelContribution[i];			 
				 System.out.println("Failure Rate "+i+" at iteration "+Iteration+":   "+failureRate[i]);
			 }
			 for (int i=0; i<numberOfPartitions;i++ ){
				 if (sumFailureRate == 0) // ALL PARTITTIONS WITH ZERO FAILURES
					 failureRate[i]=0;
				 else
					 failureRate[i] = failureRate[i]/sumFailureRate;  //normalization of phiImpr
				 System.out.println("NORMALIZED Failure Rate "+i+" at iteration "+Iteration+":   "+failureRate[i]);

				unrelContribution[i] = failureRate[i]*p_i[i];    // this is p_i * FailureRate_i  ... it represents the expected contribution to unreliabiltiy expected if we select a test case from subdomain i. The assignment of test cases is given proportionally to this value.
				 sumUnrelContribution += unrelContribution[i];			 		 
			 }

			 System.out.println("\n");
			 double sumProb =0.0;
			 for (int i=0;i<numberOfPartitions;i++){
				 if (sumUnrelContribution == 0) // ALL PARTITTIONS WITH ZERO FAILURES
					 unrelContribution[i]=0;
				 else
					 unrelContribution[i] = unrelContribution[i]/sumUnrelContribution;  //normalization of phiImpr
				
				 System.out.println("Normalized Unreliability Contribution of partition "+i+" at iteration "+Iteration+":   "+unrelContribution[i]);			
				 oldProbabilities[i]  = probabilities[i];
				 System.out.println("Old Probabilities: "+oldProbabilities[i]);
					
			//	 probabilities[i] = learningFactor * oldProbabilities[i] + (1-learningFactor)*(1-oldProbabilities[i])*unrelContribution[i];
			//	 probabilities[i] = learningFactor * oldProbabilities[i] + (1-learningFactor)*unrelContribution[i];
		//		 probabilities[i] = learningFactor * oldProbabilities[i] + (1-learningFactor)*(1-oldProbabilities[i])*(1-unrelContribution[i]);
				 probabilities[i] = learningFactor * oldProbabilities[i] + (1-learningFactor)*(1-unrelContribution[i]);

                 if (partitionFull[i]==true){ // partition full can no longer receive test cases
                     System.out.println("Partition "+i+" Full");
                     probabilities[i]=0;
                 }
				 sumProb = sumProb +probabilities[i];
			 }

		//	normalization of probabilities and UPDATE of NON ZERO PARTITIONS (an info needed to Importance sampling)
			 int NonZeroPartitions=0; //number of partitions from which at least one test case has been drawn in the previous iteration
		//	 System.out.println("\n");
			 int indexOfSingleNZP=0; // track that at least one partition had non-zero test cases in the previous iteration (useful in the extreme case decribed later)
			 for (int i=0;i<numberOfPartitions;i++){
				 probabilities[i] = probabilities[i]/sumProb;  //NORMALIZATION
				 System.out.println("Probability used to assign test cases to partition  "+i+" at iteration "+Iteration+":   "+probabilities[i]+"");
				 if (TestCasePerPartition[i]!=0) {
					 NonZeroPartitions++;
					 indexOfSingleNZP = i;
				 }
			 }	
			 System.out.println("\nNumber of partitions that received at least one test case previously "+NonZeroPartitions+"\n");
		/********CALL TO IMPORTANCE SAMPLING THAT DECIDES THE BEST NUMBER OF TEST CASES TO DISTRIBUTE IN THE NEXT ITERATION AND THE TESTS PER PARTITION (OUTPUT PARAM: TCPerPartitionNextIter)  *******/
		Double[] TCPerPartitionNextIter = new Double[numberOfPartitions];
		numberOfTestCasesNextIteration = 0;
		System.out.println("\n--- CALL TO IMPORTANCE SAMPLING ---\n");
		numberOfTestCasesNextIteration = importanceSampling(TCPerPartitionNextIter, probabilities, NonZeroPartitions, significance, desiredError, numberOfPartitions);

		System.out.println("Outcome: \nUnadjusted Number Of Test Cases for the Next Iteration: "+numberOfTestCasesNextIteration);
		for (int i=0;i<numberOfPartitions;i++)
			System.out.println("Unadjusted Number of suggested Test Cases per partition: "+TCPerPartitionNextIter[i].doubleValue());
		
   		//if (numberOfTestCasesNextIteration == 2){ // THIS CAN HAPPEN TOWARD THE END OF TESTING WHEN NonZeroPartitions=1;
		if (NonZeroPartitions == 1){	
			if (indexCurrentTest + numberOfTestCasesNextIteration > totalTests)
   				TestCasePerPartition[indexOfSingleNZP] = (totalTests - (indexCurrentTest + numberOfTestCasesNextIteration));
   			else
   				TestCasePerPartition[indexOfSingleNZP] = numberOfTestCasesNextIteration;
   		//	if all Partitions FULL  //MAANGED LATER
   			//(totalTests - indexCurrentTest); // IN TIHS CASE ASSSIGN REMAINING TESTS TO A PARTITION THAT RECEIVED SOME TESTS IN THE PREVIOUS ITEARATION
   		}
   	//	else{ 		// THE FOLLOWING IS TO MANAGE THE DISTRIBUTION OF FEW TESTS DUE TOAPPROXIMATIONS
        sumTC_overPartitions=0;
        for (int i=0;i<numberOfPartitions;i++){ // THE ACTUAL NUMBER OF TESTS IS OBTAINING BY ROUNDING THE RESULTS TCPerPartitionNextIter
            TestCasePerPartition[i] = (int)Math.floor(TCPerPartitionNextIter[i].doubleValue());
            sumTC_overPartitions = sumTC_overPartitions  +TestCasePerPartition[i];
        }
        if ((numberOfTestCasesNextIteration - sumTC_overPartitions) > 0) //IF THERE IS ANY REMAINING TEST DUE TO THE APPROXIMATION, REDISTRIBUTE IT EQUALLY (Approximately, since they could be less than number of partitions)
            for (int i=0; i< (numberOfTestCasesNextIteration - sumTC_overPartitions); i++ ){
                //for (int j=0;j<numberOfPartitions;j++)
                if (NonZeroPartitions == 1)
                    TestCasePerPartition[indexOfSingleNZP]++;
                else
                    TestCasePerPartition[i%indexOfSingleNZP]++;
            }
   	
        /**********************  SPECIAL CASES **********************/
   			int toRedistribute =0;
   			int remainingTests=0;
            if (numberOfTestCasesNextIteration >(totalTests - indexCurrentTest)){ //IF THEY ARE MORE THAN THE RESIDUAL TESTS, 
   																			//THEY SHOULD BE REMOVED. THIS HAPPENS TOWARD THE END OF TESTING 
   				System.out.println("WARNING: assigned more tests than available tests, maybe testing budget is ending");   
   				System.out.println("Total tests: "+(totalTests));
   				System.out.println("Remaining tests: "+(totalTests - indexCurrentTest));
   				remainingTests = (totalTests - indexCurrentTest);
                sumTC_overPartitions=0;
   				for (int i=0;i<numberOfPartitions;i++){
   					TestCasePerPartition[i] = (int)Math.floor((TCPerPartitionNextIter[i].doubleValue())*((double)remainingTests/(double)numberOfTestCasesNextIteration));
   					sumTC_overPartitions = sumTC_overPartitions  +TestCasePerPartition[i];
   				}
   				if ((remainingTests - sumTC_overPartitions) > 0) {//IF THERE IS STILL ANY REMAINING TEST DUE TO THE APPROXIMATION, REDISTRIBUTE IT EQUALLY (Approximately, since they could be less than number of partitions)
                    for (int i=0; i< (remainingTests - sumTC_overPartitions); i++ ){
                      for (int k=0; k<numberOfPartitions;k++){
                            if (partitionFull[k]==false){   //if (TestCasePerPartition[i] !=0)TestCasePerPartition[i]++;
                                TestCasePerPartition[k]++;
                                if (TestCasePerPartition[k]>=(domainSize[k] - sumExecutedTestCasePerPartition[k])){
                                    partitionFull[k]=true;      
                                }
                                break;
                            }
                        }
   	   				}
   				}
                for (int i=0;i<numberOfPartitions;i++) System.out.println("RESCALED number of tests pre partition: "+TestCasePerPartition[i]);           
                
   			}
             double[] tempSumExecutedTestCasePerPartition = new double[numberOfPartitions];
   			for (int i=0;i<numberOfPartitions;i++)
   				tempSumExecutedTestCasePerPartition[i] = sumExecutedTestCasePerPartition[i]; //tests executed until iteration "k-1"
   			
   			// MANAGE THE CASE OF A NUMBER OF ASSIGNED TESTS GREATER THAN ONES THAT CAN BE DONE FOR THAT PARTITION 
   			int diff=0; 
   			for (int i=0;i<numberOfPartitions;i++){
   			   
   				numberOfPotentialTests[i] = (domainSize[i] - tempSumExecutedTestCasePerPartition[i]); //all potential tests minus the number of already executed tests 
   				if (TestCasePerPartition[i] > numberOfPotentialTests[i]){  
   					System.out.println("\nWARNING: Number of Tests to execute for partition "+i+" exceeded the residual tests that can be done\n");
   					System.out.println(tempSumExecutedTestCasePerPartition[i]+" tests already executed for this partition.");
   					System.out.println(numberOfPotentialTests[i]+" tests that could still be executed.");
   					System.out.println(TestCasePerPartition[i]+" tests assigned to this partition.");
   					partitionFull[i] = true;
   					diff = (int) (TestCasePerPartition[i] -  numberOfPotentialTests[i]);
   					toRedistribute  = toRedistribute + diff; 
   					System.out.println("\n Tests to Redistribute: "+toRedistribute);
   					TestCasePerPartition[i] = TestCasePerPartition[i]-diff;
   				}
                if (TestCasePerPartition[i] == numberOfPotentialTests[i])
                    partitionFull[i] = true;
   			}
   			//REDISTRIBUTE
   			int partitionAvailable=0;
   			int[] availability = new int[numberOfPartitions];
   			for (int i=0; i<numberOfPartitions; i++ ){
   				if( numberOfPotentialTests[i] - TestCasePerPartition[i]  > 0) { // THERE IS SPACE IN THIS PARTITION
   					partitionAvailable++;
   					availability[i] = (int) (numberOfPotentialTests[i] - TestCasePerPartition[i]);  
   				}
   			}
   			if (toRedistribute!=0){
   				for (int i=0; i<numberOfPartitions; i++ ){
   					if(availability[i]>0){ 
   						if(toRedistribute<=availability[i]){ 
   							if(toRedistribute==availability[i]) partitionFull[i] = true;
   							TestCasePerPartition[i] = TestCasePerPartition[i] + toRedistribute;
   							System.out.println("\n Redistributed "+toRedistribute+" tests to partition "+i);
   							toRedistribute=0;
   							
   						}
   						else{
   							TestCasePerPartition[i] = TestCasePerPartition[i] + availability[i];
   							toRedistribute =toRedistribute - availability[i];		
   							System.out.println("\nRedistributed "+availability[i]+" tests to partition "+i);
   						}		
   					}
   				}
   				System.out.println("\n END Redistribution \n");
   			}
    
   		numberOfTestCasesNextIteration=0; // FINAL NUMBER OF TEST CASES AFTER THE ADJUSTMENTS
		for (int i=0;i<numberOfPartitions;i++) numberOfTestCasesNextIteration = (int) (numberOfTestCasesNextIteration + TestCasePerPartition[i]);
		if (numberOfTestCasesNextIteration==0){
			for (int i=0; i<numberOfPartitions; i++ ){
				if (partitionFull[i] == false){
					TestCasePerPartition[i]=1;
					numberOfTestCasesNextIteration=1;
					break;
				}	
			}
		}	
		if (numberOfTestCasesNextIteration==0){
			System.out.println("ERROR: ALL PARTITIONS ARE FULL, BUT THERE ARE STILL TESTS TO EXECUTE. POSSIBLE MISMATCH BETWEEN NUMBER OF TESTS AND PARTITION SIZE");
			System.exit(0);	
		}
		
   		System.out.println("Adjusted Number Of Test Cases for the Next Iteration: "+numberOfTestCasesNextIteration);
   		for (int i=0;i<numberOfPartitions;i++) System.out.println("TestCasePerPartition "+i+":  "+TestCasePerPartition[i]);
   		for (int i=0;i<numberOfPartitions;i++)  System.out.println("TRUE "+i+" : "+partitionFull[i]);
     System.out.println("\n");
	 for (int i=0; i<numberOfPartitions;i++ ) System.out.println("Cumulative sumExecutedTestCasePerPartition partition "+i+" at iteration "+Iteration+" : "+sumExecutedTestCasePerPartition[i]+"");

   	System.out.println("\n ---- EXECUTE TESTS BY THE PYHTON SELEXTION ALGORITHM ---\n");
   	 //ESECUZIONE PER PARTIZIONE, METTI NEL FOR
	selectedTestCasesRT.clear();//ATTENTION: THIS CAN REDO THE INTIAL SELECTION
   	failedTestCases.clear();
//	System.out.println("Working Directory = " + System.getProperty("user.dir"));
	//String 
	command= null; 
	//double
	sumOverPartitions=0;
	for (int k=0; k<numberOfPartitions;k++){
		if(TestCasePerPartition[k]!=0){
			command="python select.py "+program_name+" --ver "+version_name+" -c "+coverageCriterion+" --fm all -i "+countIteration[k]+" -p "+k+" -n "+(int)TestCasePerPartition[k]+"";		
			countIteration[k]++; 
			 System.out.println("Python Command: "+command);
			 Process p = Runtime.getRuntime().exec(command);
			 BufferedReader br = new BufferedReader(new InputStreamReader(p.getInputStream()));
			 BufferedReader stdError = new BufferedReader(new InputStreamReader(p.getErrorStream()));
			 int returnCode = p.waitFor();
			 //System.out.println("Python Script or OS Return Code: " + Integer.toString(returnCode));
			if (returnCode >= 2) {
				System.out.println("OS Error: Unable to Find File or other OS error.");
			}   	
			String str =null;  
			while (br.ready()) {
				str = br.readLine();
				System.out.println("Reading stdin...");
				System.out.println(str);
			}
		//	String str =null; 
		  if (!br.ready()){
			while (stdError.ready()) {
				str = stdError.readLine();
				System.out.println("Reading stderr...");
				System.out.println(str);
				}
			}
			
			// READ FAILED TEST CASES FROM STR   
			StringTokenizer st = new StringTokenizer(str, ", ");
				//ArrayList<Integer> faultList = new ArrayList<Integer>();
			int numberOfPerPartitionTestCases = 0;
			while(st.hasMoreTokens()){
				selectedTestCasesRT.add(Integer.parseInt(st.nextToken().trim()));			
				numberOfPerPartitionTestCases++;
			}
			System.out.println("Number of assigned Test Cases by RT to partition "+k+" :"+TestCasePerPartition[k]);			
			System.out.println("Number of selected Test Cases by RT in partition "+k+" :"+numberOfPerPartitionTestCases);			
			System.out.println("Selected Test Cases by RT :"+selectedTestCasesRT);
			
			if(numberOfPerPartitionTestCases!=TestCasePerPartition[k])
				System.exit(0);//AGGIURNA INDICI CASI DI TEST ESEGUITI E PER PARTIZIONE
			 failedTestCases = getFailedTests(previousFaultMatrixRT, selectedTestCasesRT);
			 System.out.println("FAILED TEST CASES IN ADAPTIVE RELIABILITY TESTING: "+failedTestCases);
			 if(k==0){
				 numberOfFailures[k] = failedTestCases.size();
				 sumOverPartitions = numberOfFailures[k];
			 }
			 else{
				 numberOfFailures[k] = failedTestCases.size() - sumOverPartitions;	
				 sumOverPartitions = failedTestCases.size();
				 }
		 }
		else{
			numberOfFailures[k]=0;
		}
		}
	 this.cumulativeFailedTestCases_RT.addAll(failedTestCases);
		
	this.cumulativeSelectedTestCasesRT.addAll(selectedTestCasesRT);
	 /* COMPUTE DELIVERED RELIABILITY AT ITERATION K */ 
	 this.numberDetectedFaults =0;
	 this.reliabiltiyGrowthList.clear();
	 //HashMap<Integer, ArrayList<Integer>> 
	 updatedFaultMatrixRT = updateFaultMatrix(failedTestCases,previousFaultMatrixRT); 
	 previousFaultMatrixRT = updatedFaultMatrixRT; 
	 this.reliabiltiyGrowthList_RT.addAll(this.reliabiltiyGrowthList);
	 this.cumulativeNumberFaultsRT.add(this.numberDetectedFaults);
	 if(achievedMaxReliabilityOT == true){
		 //for(int j=0; j<reliabiltiyGrowthList_RT.size();j++) this.reliabiltiyGrowthList_OT.add(this.reliabiltiyGrowthList_OT.get(this.reliabiltiyGrowthList_OT.size() - 1 ));
		 this.cumulativeNumberFaultsOP.add(0);//this.cumulativeNumberFaultsOP.get(this.cumulativeNumberFaultsOP.size()-1));	 
	 }
	
	 System.out.println("cumulativeNumberFaultsRT "+cumulativeNumberFaultsRT);	
	  reliability = computeDeliveredReliability(updatedFaultMatrixRT);
	 System.out.println("\nRELIABILITY of Adaptive Reliability Testing at iteration "+Iteration+" : "+reliability+"\n");
	// if(reliability==1) 
	//	 achievedMaxReliabilityRT = true;
	 /* COMPUTE DELIVERED RELIABILITY AT ITERATION K  - END*/
	}  // END OF if(achievedMaxReliabilityRT == false)
	
	
	/* OPERATIONAL TESTING  */ 
	

if(achievedMaxReliabilityOT == false){
	int numberOfTestCases_OT;
	if(achievedMaxReliabilityRT == true)
		 numberOfTestCases_OT = 1;//to track rel growth at each test case////(this.totalTests - indexCurrentTest);
	else
		 numberOfTestCases_OT = numberOfTestCasesNextIteration;

	reliabilityOT =0; 
	if (!previousFaultMatrixOT.isEmpty()){
		 failedTestCases_OT.clear();
		 withoutReplacement = true;
		 ArrayList<Integer> selecetdTestCasesOT = selectTestCases_OT(numberOfTestCases_OT, withoutReplacement );
		 System.out.println("SELECTED TEST CASES BY OPERATIONAL TESTING: "+selecetdTestCasesOT);
		 this.cumulativeSelectedTestCasesOT.addAll(selecetdTestCasesOT);
		 failedTestCases_OT = getFailedTests(previousFaultMatrixOT, selecetdTestCasesOT);
		 System.out.println("FAILED TEST CASES IN OPERATIONAL TESTING: "+failedTestCases_OT);
		 this.cumulativeFailedTestCases_OT.addAll(failedTestCases_OT);
		 this.numberDetectedFaults=0;
		 this.reliabiltiyGrowthList.clear();
		 HashMap<Integer, ArrayList<Integer>> updatedFaultMatrixOT = updateFaultMatrix(failedTestCases_OT,previousFaultMatrixOT);
		 previousFaultMatrixOT = updatedFaultMatrixOT; 
		 this.reliabiltiyGrowthList_OT.addAll(this.reliabiltiyGrowthList);
		 this.cumulativeNumberFaultsOP.add(this.numberDetectedFaults);
		 if(achievedMaxReliabilityRT == true){
			// for(int j=0; j<reliabiltiyGrowthList_OT.size();j++) this.reliabiltiyGrowthList_RT.add(this.reliabiltiyGrowthList_RT.get(this.reliabiltiyGrowthList_RT.size() - 1 ));
			 this.cumulativeNumberFaultsRT.add(0);//this.cumulativeNumberFaultsRT.get(this.cumulativeNumberFaultsRT.size()-1));	 
		 }
		 System.out.println("cumulativeNumberFaultsOP "+cumulativeNumberFaultsOP);	
		 reliabilityOT = computeDeliveredReliability(updatedFaultMatrixOT);	     
	 }
	 else{
		reliabilityOT = computeDeliveredReliability(previousFaultMatrixOT);
	}		 
	System.out.println("\nReliability of Operational Testing after "+(numberOfTestCases_OT+indexCurrentTest)+" test cases (iteration "+Iteration+"): "+reliabilityOT+"\n");
	if(reliabilityOT==1) achievedMaxReliabilityOT = true;
}

		 	 
		 //UPDATE INDEXES
//if((reliability<1)){
int indexCurrentTestTemp = indexCurrentTest;
		 for (int i=0; i<numberOfPartitions;i++ ){ 
			 indexCurrentTest = (int) (indexCurrentTest + TestCasePerPartition[i]); //EXECUTED TESTS
			 sumExecutedTestCasePerPartition[i] += TestCasePerPartition[i];
			 System.out.println("numberOfFailures observed for partition  "+i+":  "+numberOfFailures[i]); 
		 }
		 System.out.println("\n");
		 for (int i=0; i<numberOfPartitions;i++ ) System.out.println("Cumulative sumExecutedTestCasePerPartition partition "+i+" at iteration "+Iteration+" : "+sumExecutedTestCasePerPartition[i]+""); 
		 System.out.println("\nindexCurrentTest (executed tests) at iteration "+Iteration+": "+indexCurrentTest+" \n\n"); 
//}
if((achievedMaxReliabilityRT == true) && (achievedMaxReliabilityOT == false)){
	indexCurrentTest = indexCurrentTestTemp + 1;
	 System.out.println("\n Updated indexCurrentTest (executed tests) at iteration "+Iteration+": "+indexCurrentTest+" \n\n"); 

}
if((achievedMaxReliabilityRT == true) && (achievedMaxReliabilityOT == true)){
	indexCurrentTest = indexCurrentTestTemp;
	 System.out.println("\n Updated indexCurrentTest (executed tests) at iteration "+Iteration+": "+indexCurrentTest+" \n\n"); 

}

		 reliabilitiesSequence_OT.add(reliabilityOT);
		 reliabilitiesSequence_RT.add(reliability);
		 testCaseSequence.add(indexCurrentTest);
		 
		 System.out.println("\n ******** END OF ITERATION "+Iteration+" **********\n\n"); 
		Iteration++;
		
		if(reliability==1 && reliabilityOT==1){
		//	if(winningSet == false)
		//		returnValue = -1;
			achievedMaxReliability = true;
			}
		if(winningSet == false){
		 	if (reliability==1 && reliabilityOT<1){
			//	 returnValue = 1;
				 winningSet = true;
				 achievedMaxReliabilityRT = true;
			 }
			 if (reliabilityOT==1 && reliability<1){
			//	 returnValue = 0;
				 winningSet = true;
				 achievedMaxReliabilityOT = true;
			 }
		 }
	
    }
	
	computeInterFailureTests();
	printResults();
	if(this.progressiveTestAtRelGrowth_RT.get(this.progressiveTestAtRelGrowth_RT.size()-1) < this.progressiveTestAtRelGrowth_OT.get(this.progressiveTestAtRelGrowth_OT.size()-1))
		return 1; 
	if(this.progressiveTestAtRelGrowth_RT.get(this.progressiveTestAtRelGrowth_RT.size()-1) > this.progressiveTestAtRelGrowth_OT.get(this.progressiveTestAtRelGrowth_OT.size()-1))
		 return 0;
	if(this.progressiveTestAtRelGrowth_RT.get(this.progressiveTestAtRelGrowth_RT.size()-1) == this.progressiveTestAtRelGrowth_OT.get(this.progressiveTestAtRelGrowth_OT.size()-1))
		return -1;
	
	
	return returnValue;
	
	/*	  if(reliability==1 && reliabilityOT==1)
			 return -1;
		 if (reliability==1 && reliabilityOT<1)
			 return 1;
		 if (reliabilityOT==1 && reliability<1)
			 return 0;
		 return -2;
		 */
}

private void computeInterFailureTests(){
	//COMPUTE NUMBER OF TEST CASES BETWEEN FAILURES (i.e. AT EACH RELIABILITY GROWTH POINT)
	int indexAtRelOne = this.reliabiltiyGrowthList_OT.indexOf(1.0) ;	
	for (int index =0 ; index<=indexAtRelOne;index++){
			progressiveTestAtRelGrowth_OT.add(cumulativeSelectedTestCasesOT.indexOf(this.cumulativeFailedTestCases_OT.get(index)));
		}
	indexAtRelOne = this.reliabiltiyGrowthList_RT.indexOf(1.0) ;
	for (int index =0 ; index<=indexAtRelOne;index++){
		progressiveTestAtRelGrowth_RT.add(cumulativeSelectedTestCasesRT.indexOf(this.cumulativeFailedTestCases_RT.get(index)));
	}
}
private static int importanceSampling(Double[] TCPerPartitionNextIter, double[] probabilities, int  NonZeroPartitions, double significance, double desiredError, int numberOfPartitions){

		
		ArrayList<Double> prob = new ArrayList<Double>();
		for (int i=0; i<probabilities.length;i++)
			prob.add(probabilities[i]); 
		
		ArrayList<Double> probVectorToStoreIndexes = new ArrayList<Double>(prob);
		
		Collections.sort(prob);
		Collections.reverse(prob);
		
		
		double[] b = new double[numberOfPartitions];  
		double[] NT = new double[numberOfPartitions]; //numberof test cases per partition : temp variable
		b[0] = prob.get(0).doubleValue(); 
		for (int i=0; i<numberOfPartitions;i++)
			NT[i]=0;
		for (int i=1; i<numberOfPartitions;i++){
			b[i]=b[i-1] + prob.get(i).doubleValue();
		}
		// calcola numero di campioni con KLD

		
		//calcolo di z_(1-delta)
		NormalDistribution nd = new NormalDistribution();
		double z = nd.inverseCumulativeProbability(1-significance);
	
		int numberOfTestCaseNextIter;
		if(NonZeroPartitions>1){
			numberOfTestCaseNextIter = (int)Math.floor(((NonZeroPartitions -1)/(2*desiredError))*Math.pow((1-(2.0/(9.0*(NonZeroPartitions-1))) + Math.sqrt(2.0/(9.0*(NonZeroPartitions-1)))*z), 3));
		}
		else{
			//numberOfTestCaseNextIter = 0;
			//numberOfTestCaseNextIter = 1;
				numberOfTestCaseNextIter = (int)Math.floor(((NonZeroPartitions)/(2*desiredError))*Math.pow((1-(2.0/(9.0*(NonZeroPartitions))) + Math.sqrt(2.0/(9.0*(NonZeroPartitions)))*z), 3));
				numberOfTestCaseNextIter = (int)Math.ceil(((double)(numberOfTestCaseNextIter))/2);
			for (int i=0; i<numberOfPartitions;i++) TCPerPartitionNextIter[i] =0.0;
		//	TCPerPartitionNextIter[indexNonZero] =numberOfTestCaseNextIter; LO FACCIO NEL CLIENT
			//return numberOfTestCaseNextIter;
		}
		double[] r = new double[numberOfTestCaseNextIter+1];
	//	System.out.println("numberOfTestCaseNextIter: "+numberOfTestCaseNextIter);
		
		r[0] = Math.random()*(1.0/numberOfTestCaseNextIter);
		int i=0;
		for (int j=0; j<numberOfTestCaseNextIter;j++){
			while (r[j]>b[i]){
				i++;
			}
			NT[i]++;
			r[j+1] = r[j] + 1.0/numberOfTestCaseNextIter; 
		}
			
		//%riordinare secondo l'ordine iniziale
//		boolean canDecreaseFromOthers = false;
	//	if (numberOfTestCaseNextIter>numberOfPartitions)
			
			
		int index; 
//		int paritionsWithZeroTC = 0;
		for (int k=0; k<prob.size();k++){			
			index = prob.indexOf(probVectorToStoreIndexes.get(k));
			TCPerPartitionNextIter[k] = NT[index];
//			if (TCPerPartitionNextIter[k]==0 && probabilities[k]>0 && canDecreaseFromOthers ){
	//			TCPerPartitionNextIter[k]++;
		//		paritionsWithZeroTC++; 
			//}
//			DEVE ANCHE ESSERE IL CASO IN CUI POSSO TOGLIERE AGLI ALTRI..AGLI.
	//		OSSIA NUMERO TOTALE DI CASI DI TEST ALMENO MAGIORE DEL NUMERO DI PARTIZIONI
		//	E PARTIZIONI DA CUI TOLGO ALMENO CON PIU DI 1 CASO DI TEST
		}

//		int indexPart;
//		indexPart = numberOfPartitions-1;
//		double max=0;
		
//		for (int i=0; i< paritionsWithZeroTC; i++ ){
//			max=0;
//			for (int k=0; k<numberOfPartitions;k++){
//				if(max<TCPerPartitionNextIter[k]) {
//					max=TCPerPartitionNextIter[k];
//					indexPart=k;
//				}	
//			}		
//			TestCasePerPartition[indexPart]--;
//		  }
		
		//STAMPO
		for (int k=0; k<numberOfPartitions;k++){
		//	System.out.println("Unordered Number of TC pre Partition: "+NT[k]);
	//		System.out.println("Final Number of TC pre Partition: "+TCPerPartitionNextIter[k]);
		}
		return numberOfTestCaseNextIter;
		
	}

private static HashMap<Integer,ArrayList<Integer>> loadFaultMatrix(Path file){
	
	HashMap<Integer,ArrayList<Integer>> matrix = new HashMap<Integer,ArrayList<Integer>>();

	// READ AND PARSE FILE
	ArrayList<String> linesList = new ArrayList<String>();
	
	Scanner scan = null;
	try {
		scan = new Scanner(file);
	} catch (IOException e) {
		System.out.println("Error in reading the fault matrix file");
		e.printStackTrace();
	}
	String[] lines = null ;
	 while(scan.hasNext())
	        lines = scan.nextLine().split("],");
	 
	
	 lines[0] = lines[0].replace("{", " ");   // ADJUST FIRST STRING
	 lines[lines.length-1] = lines[lines.length-1].substring(0, lines[lines.length-1].length()-2); // ADJUST LAST STRING
			 
	 for(int i=0;i<lines.length;i++) // PUT THE STRING[] IN ARRAY_LIST TO MANAGE MORE EASILY
		 linesList.add(lines[i]);
	 Iterator<String> it = linesList.iterator();
	 StringTokenizer st, stFaultList;
	 String line;	
	 Integer ID_TestCase = null;
	while (it.hasNext()){
			line = new String(it.next());
			st = new StringTokenizer(line);
			ArrayList<Integer> faultList = new ArrayList<Integer>();
			
		while(st.hasMoreTokens()){
				String token = st.nextToken(":").trim();
				if(token.contains("[")){
				//	System.out.println("I AM HERE "+token.toString());
					stFaultList = new StringTokenizer(token.substring(1, token.length()), ", ");
					while(stFaultList.hasMoreTokens())
						faultList.add(Integer.parseInt(stFaultList.nextToken().trim()));
				}
				else{	
					ID_TestCase = Integer.parseInt(token); 
				}					
			}
			matrix.put(ID_TestCase, faultList);
			
	//		for (int j=0; j<matrix.get(ID_TestCase).size();j++) System.out.println("Fault Matrix - Faults: "+matrix.get(ID_TestCase).get(j));
		}	
	System.out.println("Fault Matrix: "+matrix);
	System.out.println("Fault Matrix size (number of test cases activating at least one fault): "+matrix.size());
	return matrix;
}

private  HashMap<Integer,Double> computeOccurrenceProbability(int[] domainSize,int totalTests, double[] probabilities){
	
	HashMap<Integer,Double> occurrenceProb = new HashMap<Integer,Double>();
	
	int IDTestCase = 1; //ATTENTION: ASSUMING THAT TEST CASES ARE INDEXED  IN A PROGRESSIVE WAY 
		for (int j=0; j< domainSize.length; j++){
			for (int i=0; i< domainSize[j]; i++){
				occurrenceProb.put(IDTestCase, (probabilities[j]/domainSize[j]));
				IDTestCase++;
				
			}
		}
		System.out.println("Occurrence probability (assumed equal distribution within each domain): "+occurrenceProb);
		double sum=0;
		for (int k= 1;k<=occurrenceProb.size();k++){
			sum = sum+occurrenceProb.get(k); 		
			}
		System.out.println("Occurrence probability size (should be equal to the number of test cases): "+occurrenceProb.size());
	return occurrenceProb;
}

private  HashMap<Integer,ArrayList<Integer>> updateFaultMatrix(ArrayList<Integer> failedTestCases, HashMap<Integer,ArrayList<Integer>> faultMatrix){
	HashMap<Integer,ArrayList<Integer>>  newFaultMatrix = new HashMap<Integer,ArrayList<Integer>>(faultMatrix);
	ArrayList<Integer> faultsToRemove = new ArrayList<Integer>();
	
	int listEmpty=0;
	int j=0;
	this.numberDetectedFaults = 0;
	if(!failedTestCases.isEmpty()){
		//System.out.println("failedTestCases.size "+failedTestCases.size());
		for(int i=0; i<failedTestCases.size();i++){
			// ASSUMING TO REMOVE ALL THE FAULTS DETECTED BY A TEST CASE
			faultsToRemove = newFaultMatrix.get(failedTestCases.get(i));
			newFaultMatrix.remove(failedTestCases.get(i));  // TOGLI TEST CASE CORRISPONDENTI a test fallito (RIMUOVO TUTTI I FAULT)
			if(faultsToRemove!=null){
				this.numberDetectedFaults += faultsToRemove.size();
				//System.out.println("numberDetectedFaults in FM "+numberDetectedFaults);

				for (int k= 0; k<faultsToRemove.size();k++) System.out.println("Faults #: "+faultsToRemove.get(k));
				for(Map.Entry<Integer, ArrayList<Integer>> entry : newFaultMatrix.entrySet()){
					for (int index=0; index<faultsToRemove.size();index++){ 
						int indexTemp =entry.getValue().indexOf(faultsToRemove.get(index));
						if(indexTemp != -1) entry.getValue().remove(indexTemp);
					}
			//		System.out.println(entry.getKey() + " = " + entry.getValue());
				}
				//updatedFaultsList = new ArrayList<Integer>(updatedFaultsListTemp);
				//newFaultMatrix.put(key, updatedFaultsList);
			}
			else
				j++;
			//RELIABILTIY AFTER EACH FAILING TEST CASE
			System.out.println("Reliabiltiy after each failing test case: "+computeDeliveredReliability(newFaultMatrix));
			//RECORD THE TEST CASE AND THE NUMBER OF TESTS EXECUTED SO FAR AT THE CALLER SITE
			//if (reliabiltiyGrowthList.get(reliabiltiyGrowthList.size()-1)!=1)
			this.reliabiltiyGrowthList.add(computeDeliveredReliability(newFaultMatrix));
		}
	//	System.out.println("Duplicates: "+j);
		
		//REMOVE EMPTY ELEMENTS (WHEN THE TEST CASE DOES NO LONGER DETECT FAULTS)
//		System.out.println("UPDATED FAULT MATRIX SIze before: "+newFaultMatrix.size());

		listEmpty=0;
		Iterator iterator = newFaultMatrix.entrySet().iterator();
		while(iterator.hasNext()){
			 Map.Entry<Integer, ArrayList<Integer>> pair = (Entry<Integer, ArrayList<Integer>>)iterator.next();
	//	     System.out.println(pair.getKey() + " = " + pair.getValue());
			if(pair.getValue().isEmpty()){
				listEmpty++;
				iterator.remove(); 
			}
		}
	
	}
	System.out.println("\nUPDATED FAULT MATRIX: "+newFaultMatrix);
//	System.out.println("Number of test cases that have been corrected: "+(listEmpty));
	System.out.println("INITIAL FAULT MATRIX SIZE: "+faultMatrix.size());
	System.out.println("UPDATED FAULT MATRIX SIZE: "+newFaultMatrix.size());
	return newFaultMatrix;
}

private  double computeDeliveredReliability(HashMap<Integer,ArrayList<Integer>> newFaultMatrix){
	
	//CLEAN THE FAULT MATRIX FROM POSSIBLE EMPTY ELEMENTS (TEST CASES WITH NO FAULT LIST ATTACHED)
	int listEmpty=0;
	Iterator it = newFaultMatrix.entrySet().iterator();
	while(it.hasNext()){
		 Map.Entry<Integer, ArrayList<Integer>> pair = (Entry<Integer, ArrayList<Integer>>)it.next();
		if(pair.getValue().isEmpty()){
			listEmpty++;
			it.remove(); 
		}
	}
	
	double unrel = 0.0;
	Iterator<Integer> iterator =  newFaultMatrix.keySet().iterator();
	int counter=0;
	while(iterator.hasNext()){
				unrel = unrel +  occurrenceProbability.get(iterator.next()).doubleValue();		
				counter++;
		}	
	double deliveredReliability =(1- unrel); 
	return deliveredReliability;
}

private void printResults(){
	
	// STAMPA LIST OF EXECUTED TEST CASES
	pw.println("\nList of Executed Test Cases by Reliability Testing: "+this.cumulativeSelectedTestCasesRT);
	pw.println("List of Executed Test Cases by Operational Testing: "+this.cumulativeSelectedTestCasesOT);
	
	pw.println("\nList of Failed Test Cases by Reliability Testing: "+this.cumulativeFailedTestCases_RT);
	pw.println("List of Failed Test Cases by Operational Testing: "+this.cumulativeFailedTestCases_OT);
	
	pw.println("\nNumber of Executed Test Cases at each iteration \n"+this.testCaseSequence);
	pw.println("Reliability growth values of Operational Testing at each iteration \n"+this.reliabilitiesSequence_OT);
	pw.println("Reliability growth values of Adaptive reliability Testing at each iteration \n"+this.reliabilitiesSequence_RT);
	
	pw.println("\nNumber of Faults detected by Operational Testing at each iteration \n"+this.cumulativeNumberFaultsOP);
	pw.println("Number of Faults detected by Adaptive reliability Testing at each iteration \n"+this.cumulativeNumberFaultsRT);
	
	pw.println("\n\nReliability growth of RT at each failed test case (and corresponding correction of faults): "+this.reliabiltiyGrowthList_RT);
	pw.println("\n Number of already executed tests before the i^th failures (RT) - 'i=1:NumberOfFailures' is the index of the following vector: "+progressiveTestAtRelGrowth_RT);
	
	
	pw.println("\n\nReliability growth of OT at each failed test case (and corresponding correction of faults): "+this.reliabiltiyGrowthList_OT);
	pw.println("\n Number of executed tests untile the i^th test failures (OT) 'i=1:NumberOfFailures' is the index of the following vector: "+progressiveTestAtRelGrowth_OT);
	
	
	pw.println("\n\n");
	pw.flush();
//	pw.close();
	
	// PRINT RESULTS ON CSV FILES
//	csv_w.println("");
	
}


public String getCsv_string(int fileType, int repetitionInfo) {
	
	if (repetitionInfo<0)
		throw new IllegalArgumentException("Wrong information on which 'Repetition' is being run. The repetition ID must be >= 0");
	
	StringBuilder sb = new StringBuilder();
	if(repetitionInfo==0){
		sb.append("repetition");
		sb.append(",");
	}
	switch(fileType){
	case 0: //PER ITERATION INFORMATION
		if(repetitionInfo==0){
			sb.append("iteration");
			sb.append(",");
			sb.append("nCumulativeExecutedTC");
			sb.append(",");
			sb.append("reliabilityGrowthRT");
			sb.append(",");
			sb.append("reliabilityGrowthOT");
			sb.append(",");
			sb.append("nFaultsRTperIteration");
			sb.append(",");
			sb.append("nFaultsOTperIteration");
			sb.append("\n");
		}

		System.out.println("TEST CASE SEQUENCE "+testCaseSequence);
		for(int i=0;i<this.testCaseSequence.size();i++){
			sb.append(repetitionInfo);
			sb.append(",");
			sb.append(i);
			sb.append(",");
			sb.append(this.testCaseSequence.get(i));
			sb.append(",");
			sb.append(this.reliabilitiesSequence_RT.get(i));
			sb.append(",");
			sb.append(this.reliabilitiesSequence_OT.get(i));
			sb.append(",");
			sb.append(this.cumulativeNumberFaultsRT.get(i));
			sb.append(",");
			sb.append(this.cumulativeNumberFaultsOP.get(i));
			sb.append("\n");			
		}
			
		break;
	case 1: //PER REPETITION INFORMATION
		if(repetitionInfo==0){
			sb.append("listExecutedTC_RT");
			sb.append(",");
			sb.append("listExecutedTC_OT");
			sb.append(",");
			sb.append("listFailedTC_RT");
			sb.append(",");
			sb.append("listFailedTC_OT");
			sb.append("\n");
		}
		System.out.println("cumulativeSelectedTestCasesRT "+cumulativeSelectedTestCasesRT);
		System.out.println("suze rt "+cumulativeSelectedTestCasesRT.size());
		System.out.println("cumulativeSelectedTestCasesOT "+cumulativeSelectedTestCasesOT);
		System.out.println("size ot "+cumulativeSelectedTestCasesOT.size());		
		System.out.println("cumulativeFailedTestCases_RT "+cumulativeFailedTestCases_RT);
		System.out.println("cumulativeFailedTestCases_OT "+cumulativeFailedTestCases_OT);
		System.out.println("cumulativeFailedTestCases_RT size "+cumulativeFailedTestCases_RT.size());
		System.out.println("cumulativeFailedTestCases_OT size "+cumulativeFailedTestCases_OT.size());
		int biggerSize,smallerSize =  0;
		if(this.cumulativeSelectedTestCasesRT.size()>this.cumulativeSelectedTestCasesOT.size()){
			biggerSize = this.cumulativeSelectedTestCasesRT.size();
			smallerSize = this.cumulativeSelectedTestCasesOT.size();
			}
		else{
			biggerSize = this.cumulativeSelectedTestCasesOT.size();
			smallerSize = this.cumulativeSelectedTestCasesRT.size();
		}
		for(int i=0;i<biggerSize;i++){
			sb.append(repetitionInfo);
			sb.append(",");
			if(i<smallerSize){
				sb.append((this.cumulativeSelectedTestCasesRT.get(i)).toString());
				sb.append(",");
				sb.append((this.cumulativeSelectedTestCasesOT.get(i)).toString());
				sb.append(",");
				if(this.cumulativeFailedTestCases_RT.contains(this.cumulativeSelectedTestCasesRT.get(i)))
					sb.append("yes");
				else
					sb.append("no");
				sb.append(",");
				
				if(this.cumulativeFailedTestCases_OT.contains(this.cumulativeSelectedTestCasesOT.get(i)))
					sb.append("yes");
				else
					sb.append("no");
			}
			else{
				if(this.cumulativeSelectedTestCasesRT.size()==biggerSize){
					sb.append((this.cumulativeSelectedTestCasesRT.get(i)).toString());
					sb.append(",");
					sb.append("-1");
					sb.append(",");
					if(this.cumulativeFailedTestCases_RT.contains(this.cumulativeSelectedTestCasesRT.get(i)))
						sb.append("yes");
					else
						sb.append("no");
					sb.append(",");
					
					sb.append("N/A");
				}
				if(this.cumulativeSelectedTestCasesOT.size()==biggerSize){
					sb.append("-1");
					sb.append(",");
					sb.append((this.cumulativeSelectedTestCasesOT.get(i)).toString());
					sb.append(",");
					
					sb.append("N/A");
					sb.append(",");
					if(this.cumulativeFailedTestCases_OT.contains(this.cumulativeSelectedTestCasesOT.get(i)))
						sb.append("yes");
					else
						sb.append("no");
				}
			}
			
		
			sb.append("\n");			
		}
		break;	
	case 2: 
		if(repetitionInfo==0){
			sb.append("relGrowthAtEachFailedTest_RT");
			sb.append(",");
			sb.append("relGrowthAtEachFailedTest_OT");
			sb.append("\n");
		}
		biggerSize = smallerSize =  0;
		if(this.reliabiltiyGrowthList_RT.size()>this.reliabiltiyGrowthList_OT.size()){
			biggerSize = this.reliabiltiyGrowthList_RT.size();
			smallerSize = this.reliabiltiyGrowthList_OT.size();
			}
		else{
			biggerSize = this.reliabiltiyGrowthList_OT.size();
			smallerSize = this.reliabiltiyGrowthList_RT.size();
		}
		for(int i=0;i<biggerSize;i++){
			sb.append(repetitionInfo);
			sb.append(",");
			if(i<smallerSize){
				sb.append((this.reliabiltiyGrowthList_RT.get(i)).toString());
				sb.append(",");
				sb.append((this.reliabiltiyGrowthList_OT.get(i)).toString());
			}
			else{
				if(this.reliabiltiyGrowthList_RT.size()==biggerSize){
					sb.append((this.reliabiltiyGrowthList_RT.get(i)).toString());
					sb.append(",");
					sb.append("-1");
				}
				if(this.reliabiltiyGrowthList_OT.size()==biggerSize){
					sb.append("-1");
					sb.append(",");
					sb.append((this.reliabiltiyGrowthList_OT.get(i)).toString());
				}
			}
			
			sb.append("\n");			
		}
		break;	

	case 3: 
		if(repetitionInfo==0){
			sb.append("numExecutedTestsBeforeFailure_RT");
			sb.append(",");
			sb.append("numExecutedTestsBeforeFailure_OT");
			sb.append("\n");
		}
		biggerSize = smallerSize =  0;
		if(this.progressiveTestAtRelGrowth_RT.size()>this.progressiveTestAtRelGrowth_OT.size()){
			biggerSize = this.progressiveTestAtRelGrowth_RT.size();
			smallerSize = this.progressiveTestAtRelGrowth_OT.size();
			}
		else{
			biggerSize = this.progressiveTestAtRelGrowth_OT.size();
			smallerSize = this.progressiveTestAtRelGrowth_RT.size();
		}
		for(int i=0;i<biggerSize;i++){
			sb.append(repetitionInfo);
			sb.append(",");
			if(i<smallerSize){
				sb.append((this.progressiveTestAtRelGrowth_RT.get(i)).toString());
				sb.append(",");
				sb.append((this.progressiveTestAtRelGrowth_OT.get(i)).toString());
			}
			else{
				if(this.progressiveTestAtRelGrowth_RT.size()==biggerSize){
					sb.append((this.progressiveTestAtRelGrowth_RT.get(i)).toString());
					sb.append(",");
					sb.append("-1");
				}
				if(this.progressiveTestAtRelGrowth_OT.size()==biggerSize){
					sb.append("-1");
					sb.append(",");
					sb.append((this.progressiveTestAtRelGrowth_OT.get(i)).toString());
				}
			}
			sb.append("\n");
			}

		break;
	
	default:
		throw new IllegalArgumentException("Invalid csv file type required (0 or 1 allowed): "+fileType);
	}
		 
	return sb.toString();
}


private  ArrayList<Integer> getFailedTests(HashMap<Integer,ArrayList<Integer>> faultMatrix,  ArrayList<Integer> selecetdTests){

	ArrayList<Integer> failedTestCases = new ArrayList<Integer>();
	for(int index=0;index<selecetdTests.size();index++){
		if (faultMatrix.containsKey(selecetdTests.get(index))){
			failedTestCases.add(selecetdTests.get(index));			
		}
	}
	return failedTestCases;
}


private  ArrayList<Integer> selectTestCases_OT(int numberOfTestCases_OT, boolean withoutReplacement){
	
	ArrayList<Integer> selectedTests = new ArrayList<Integer>(); 
	// BUILD CUMULATIVE VECTOR
	// WITHOUT REPLACEMENT: BUILD THE CUMULATIVE VECTOR ON POTENTIAL LIST OF TESTS NOT ON AL TESTS
	ArrayList<Double> cumulativePVector = new ArrayList<Double>(); // new double[potentialTestCases_OT.size()];
	int[] indexVextor = new int[potentialTestCases_OT.size()];
	System.out.println("potentialTestList "+potentialTestCases_OT.size());
	for(int i=0; i<potentialTestCases_OT.size();i++){
		if (i==0) cumulativePVector.add(occurrenceProbability.get(potentialTestCases_OT.get(i)));
		else cumulativePVector.add(cumulativePVector.get(i-1)+ occurrenceProbability.get(potentialTestCases_OT.get(i)));
	//	System.out.println("cumulativePVector(i) "+cumulativePVector.get(i));
	}
	
	//EXTRACT RANDOM NUMBER to select test case
	int indexCurrentTest=0;
	int selectedTest=-1;
	while(indexCurrentTest < numberOfTestCases_OT){
		double rand = Math.random()*cumulativePVector.get(cumulativePVector.size()-1);
		for(int index =0; index < cumulativePVector.size();index++){
			if (rand <= cumulativePVector.get(index)) {
				selectedTest = potentialTestCases_OT.get(index);
		//		System.out.println("selectedTest "+selectedTest);		
				break;
				}
		}
		
		if (withoutReplacement == true){
			if(selectedTests.contains(selectedTest)){
				System.out.println("OT error: selected duplicate test case"+selectedTest);
				System.exit(0);
			}
			int TCtoRemove = this.potentialTestCases_OT.indexOf(selectedTest);
		//	System.out.println("removed index "+TCtoRemove);
		//	System.out.println("removed TC "+potentialTestCases_OT.get(TCtoRemove));
			
			potentialTestCases_OT.remove(TCtoRemove);
		//UPDATE CUMULATIVE VECTOR
		//	System.out.println("size ptc "+potentialTestCases_OT.size());
			cumulativePVector.clear();
			for(int i=0; i<potentialTestCases_OT.size();i++){
				if (i==0) cumulativePVector.add(occurrenceProbability.get(potentialTestCases_OT.get(i)));
				else cumulativePVector.add(cumulativePVector.get(i-1)+ occurrenceProbability.get(potentialTestCases_OT.get(i)));
				}
			}
		selectedTests.add(selectedTest);
		indexCurrentTest++;
		}
	return selectedTests;
}




public PrintWriter getCsv_w() {
	return csv_w;
}

public PrintWriter getPw() {
	return pw;
}


public PrintWriter getCsv_w_2() {
	return csv_w_2;
}

public PrintWriter getCsv_w_4() {
	return csv_w_4;
}

public PrintWriter getCsv_w_5() {
	return csv_w_5;
}

		
}