import os
import sys
import logging
import time
import subprocess
from pickle import dump, load

FIRST_TC_NUMBER = 1  # 0 or 1
TC_NAME_PATTERN = None  # 'TC{}'

def run_cmd(command, timeout=None, cwd=None, stdout=None):
    if stdout == None:
        stdout=subprocess.PIPE
    if not cwd:
        logging.debug('running command in the current working dir')
        process = subprocess.Popen(command, stdout=stdout, stderr=subprocess.PIPE)
    else:
        logging.debug('running command in a different working dir: {}'.format(cwd))
        process = subprocess.Popen(command, stdout=stdout, stderr=subprocess.PIPE, cwd=cwd)
    exit_code = process.wait()    
    response = process.communicate()    
    logging.debug('exit_code: {}'.format(exit_code))
    for r in response:
        if r != '':
            logging.debug('reponse: {}'.format(r))
    return exit_code, response

def get_individual_tc_coverage(coverage_folder, cov_info_files, format_=None, entity=None):
    coverage_info = {}
    for cov_info_file in cov_info_files:
        tc_number = int(cov_info_file.split('.')[0])
        if FIRST_TC_NUMBER == 0:
            tc_number -= 1
        test_case_name = tc_number
        if TC_NAME_PATTERN:
            tc_number = str(tc_number).zfill(3)
            test_case_name = TC_NAME_PATTERN.format(tc_number)

        cov_info_file_full_path = os.path.join(coverage_folder, cov_info_file)
        #print 'cov_info_file: ', cov_info_file_full_path
        if "idlelib" in sys.modules:  # running in idle
            print 'idlelib!!!'
            logging.info('capturing coverage information for test case #{:>5}'.format(test_case_name))
        else:
            sys.stdout.write('capturing coverage information for test case #{:>5}\r'.format(test_case_name))
            sys.stdout.flush()
            time.sleep(0.1)
        tc_coverage = get_coverage_info(cov_info_file_full_path, format_, entity)
        if entity and len(format_) == 1:
            coverage_info[test_case_name] = tc_coverage['{}_{}'.format(format_[0], entity)]
        else:
            coverage_info[test_case_name] = tc_coverage
        
    return coverage_info


def get_coverage_info(cov_info_file, format_=[], entity=None): 
    with open(cov_info_file, 'r') as file_:
        lines = file_.readlines()

        source_files = [line.replace('\n', '').split(':')[1].split('/')[-1] for line in lines if line.startswith("SF:")]
        if len(source_files) != 1:
            with open('coverage_attention.txt', 'a') as f:
                f.write('cov_info_file: {} | source_files: {}'.format(cov_info_file, source_files))

        #=======================================================================
        # Lines
        #=======================================================================
        aux_lines = [line.replace('\n', '').split(':')[1] for line in lines if line.startswith("DA:")]        
        #reachable_lines = [line.split(',')[0] for line in aux_lines]
        #covered_lines = [line.split(',')[0] for line in aux_lines if line.split(',')[1] != '0']        
        #line_detailed = [(e.split(',')[0], e.split(',')[1]) for e in aux_lines]
        reachable_line, covered_line, detailed_line, coveredcount_line = [], [], [], []
        for line in aux_lines:
            id_ = line.split(',')[0]
            count = line.split(',')[1]
            try:
                count = int(count)
            except ValueError:
                count = 0 
            reachable_line.append(id_)
            if count > 0:
                covered_line.append(id_)
                coveredcount_line.append((id_, count))
            detailed_line.append((id_, count))           
        #=======================================================================
        # Functions
        #=======================================================================
        aux_functions = [line.replace('\n', '').split(':')[1] for line in lines if line.startswith("FNDA")]
        #reachable_functions2 = [line.split(',')[1] for line in aux_functions]
        #covered_functions2 = [line.split(',')[1] for line in aux_functions if line.split(',')[0] != '0']
        #function_detailed2 = [(e.split(',')[1], e.split(',')[0]) for e in aux_functions]        
        reachable_function, covered_function, detailed_function, coveredcount_function = [], [], [], []
        for line in aux_functions:
            id_ = line.split(',')[1]
            count = line.split(',')[0]
            try:
                count = int(count)
            except ValueError:
                count = 0 
            reachable_function.append(id_)
            if count > 0:
                covered_function.append(id_)
                coveredcount_function.append((id_, count))
            detailed_function.append((id_, count))         
        #=======================================================================
        # Branches
        #=======================================================================
        reachable_branch, covered_branch, detailed_branch, coveredcount_branch = [], [], [], []
        for line in lines:
            count += 1
            if line.startswith("BRDA"):
                aux = line.replace('\n', '').split(':')[1].split(',')
                branch_name = '{line_number}_{block_number}_{branch_number}'.format(line_number=aux[0], block_number=aux[1], branch_number=aux[2])
                branch_count = '{branch_number}'.format(branch_number=aux[3])
                reachable_branch.append(branch_name)            
                try:
                    branch_count = int(branch_count)
                except ValueError:
                    branch_count = 0 
                if branch_count > 0:
                    covered_branch.append(branch_name)
                    coveredcount_branch.append((branch_name, branch_count))
                detailed_branch.append((branch_name, branch_count))
        #=======================================================================
        coverage_info = {}
        if 'reachable' in format_:
            coverage_info['reachable_line']     = reachable_line
            coverage_info['reachable_function'] = reachable_function
            coverage_info['reachable_branch']   = reachable_branch
        if 'covered' in format_:
            coverage_info['covered_line']       = covered_line
            coverage_info['covered_function']   = covered_function
            coverage_info['covered_branch']     = covered_branch
        if 'detailed' in format_:
            coverage_info['detailed_line']      = detailed_line
            coverage_info['detailed_function']  = detailed_function
            coverage_info['detailed_branch']    = detailed_branch
        if 'coveredcount' in format_:
            coverage_info['coveredcount_line']      = coveredcount_line
            coverage_info['coveredcount_function']  = coveredcount_function
            coverage_info['coveredcount_branch']    = coveredcount_branch
        #if entity and len(format_) == 1:
            #return                                
        #=======================================================================
        # if entity:
        #     tobedeleted = [k for k in coverage_info if not k.endswith(entity)]
        #     for key in tobedeleted:
        #         del(coverage_info[key])          
        #=======================================================================
        return coverage_info


def collect_coverage_info(version_dir, cov_info_files=None, format_=None, entity=None, dump_data=True, split_dict=True):
    coverage_info_file = os.path.join(version_dir, 'coverage_info.pickle')
    if entity:
        coverage_info_file = os.path.join(version_dir, 'coverage_info_{}.pickle'.format(entity))
    try:
        with open(coverage_info_file, 'r') as f:
            coverage_info = load(f)
        logging.info('{} already exists!'.format(coverage_info_file))
    except IOError:
        logging.info('{} does not exist yet!'.format(coverage_info_file))
        coverage_folder = os.path.join(version_dir, 'individual_test_coverage')
        if not cov_info_files:
            cov_info_files = [file_ for file_ in os.listdir(coverage_folder)
                              if os.path.isfile(os.path.join(coverage_folder, file_))
                              and file_.endswith('.info')]
            cov_info_files.sort()
        coverage_info = get_individual_tc_coverage(coverage_folder, cov_info_files, format_, entity)
        logging.info('dumping coverage_info dict into pickle file...')

        #=======================================================================
        # from klepto.archives import dir_archive
        # dir_ = os.path.join(version_dir, 'demo')
        # demo = dir_archive(dir_, coverage_info, serialized=True)
        # demo.dump()
        # del(demo) 
        #=======================================================================

        #=======================================================================
        # split coverage dict per criteria
        #=======================================================================
        if dump_data and not split_dict:
            with open(coverage_info_file, 'w') as f:
                dump(coverage_info, f)
                logging.info('{} has been created!'.format(coverage_info_file))                           
        elif dump_data and split_dict and format_ == ['coveredcount']:
            for entity in ['function','line','branch']:
                aux = {}
                for tc in coverage_info:
                    aux[tc] = coverage_info[tc]['coveredcount_{}'.format(entity)]                
                outfile = os.path.join(version_dir, 'coverage_info_{}.pickle'.format(entity))   
                with open(outfile, 'w') as f:
                    dump(aux, f)
                    logging.info('{} has been created!'.format(outfile)) 
        else:
            assert False, "Needs to be implemented!"  
    
    return coverage_info


def compare_multiple_info_files(list_, entity):
    aux = []
    key = 'covered_{}'.format(entity)
    for file_ in list_:
        cov_info = get_coverage_info(file_, format_=['covered'])        
        aux.append(set(cov_info[key]))
        print file_, len(cov_info[key]), sorted(cov_info[key])
    
    inter = set.intersection(*map(set,aux))
    union = set.union(*map(set,aux))
    print 'inter: ', len(inter), sorted(inter)
    print 'union: ', len(union), sorted(union) 
    return inter, union   
    #===========================================================================
    # for set_ in aux:
    #     essential = set.difference(set_, inter)
    #     print 'essential: ', len(essential), sorted(essential)
    #     
    # diff = set.difference(union, inter)
    # print 'union - inter: ', len(diff), sorted(diff)
    #===========================================================================
    

def merge_lcov_files(lcov_files_list, outfile):
    aux_str = ''.join('-a {} '.format(file_) for file_ in lcov_files_list)
    lcov_cmd = 'lcov --rc lcov_branch_coverage=1 {} -o {}'.format(aux_str, outfile)
    logging.debug('lcov_cmd: {}'.format(lcov_cmd))
    run_cmd(lcov_cmd.split())
    # Makes sure info file has been created
    count = 0
    while not os.path.exists(outfile):      
        msg = "Waiting for info file to be created: {}s\r".format(count)
        #sys.stdout.write(msg)
        #sys.stdout.flush()        
        count += 0.1
        time.sleep(0.1)
        if count > 5.0:
            logging.info('Info file could not be created at: {}. List len: {}'.format(outfile, len(lcov_files_list)))
            logging.info('aux_str: {}'.format(aux_str))
            logging.info('lcov_cmd: {}'.format(lcov_cmd))
            break

def merge_lcov_files_step(lcov_files_list, outdir, step):
    count = 0
    for i in range(0, len(lcov_files_list), step):
        interval = lcov_files_list[i:i+step]
        if count != 0 : # there is at least one "accumulated" coverage info file
            last_acc_file = os.path.join(outdir, 'acc{}.info'.format(i))
            interval.append(last_acc_file)
            logging.debug('last_acc_file: {}'.format(last_acc_file))
            logging.debug('interval: {}'.format(interval))             
        logging.debug('i: {}, interval:{}'.format(i, interval))
        logging.info('capturing coverage information after {:>4} test cases'.format(count + step))
        outfile = os.path.join(outdir, 'acc{}.info'.format(count + step))
        merge_lcov_files(interval, outfile)
        count += step

    
if __name__ == '__main__':
    #===========================================================================
    #===========================================================================
    #  SUBJECT_NAME = 'sed'
    #  os.chdir(os.path.join('../../subjects/', SUBJECT_NAME))
    #  versions = ['v6', 'v7']
    # # #versions = [f for f in os.listdir('.') if os.path.isdir(f) and f.startswith('v') and len(f) == 2]
    #  for version in versions:
    #      coverage_info = collect_coverage_info(version)
    #      for k, v in coverage_info.iteritems():
    #          print 'TC: {}'.format(k),
    #          for i, j in coverage_info[k].iteritems():
    #              print i, len(j),
    #          print ' '
    #===========================================================================
    #===========================================================================
    #cov_info_file = 'coverage_1.info'
    #coverage_info = get_coverage_info(cov_info_file, returnAll=False)
    
    #get_individual_tc_coverage('./triangle/coverage_info', returnAll=False)
    logging.basicConfig(level=logging.DEBUG)
    #cov_info_file = "/home/breno/DevWorkspace/icseexp/subjects/grep/v3/random_ops/6/acc_2_31_1133.info"
    #get_coverage_info(cov_info_file, returnAll=True)
    dir_ = "/home/breno/DevWorkspace/icseexp/subjects/grep/v3_/individual_test_coverage"
    failing = [20, 272, 274, 692, 756]
    passing = [tc for tc in range(1, 810) if tc not in failing]
    #l = os.listdir(dir_)[:50]
    #print len(l), l
    #f_ = [os.path.join(dir_, i) for i in l]
    f_ = [os.path.join(dir_, '{}.info'.format(tc)) for tc in failing]
    p_ = [os.path.join(dir_, '{}.info'.format(tc)) for tc in passing]
    #print f_
    f_inter, f_union = compare_multiple_info_files(f_, 'function')
    for tc in p_:
        cov_info = get_coverage_info(tc, format_=['covered'])['covered_function']
        #new_inter = set.intersection(set(cov_info), (f_inter))
        #f_inter = new_inter
        f_inter.difference_update(set(cov_info))
        print len(f_inter)
    #inter, union = compare_multiple_info_files(p_, 'function')
    
    