import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.PrintStream;
import java.io.PrintWriter;
import java.io.StringWriter;
import java.nio.charset.StandardCharsets;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Scanner;
import java.util.StringTokenizer;

import org.apache.commons.math3.distribution.*;

public class main {

	//private static HashMap<Integer, ArrayList<Integer>> faultMatrix;
	private static HashMap<Integer, ArrayList<Integer>> previousFaultMatrixRT;
	private static HashMap<Integer, ArrayList<Integer>> previousFaultMatrixOT;
	private static HashMap<Integer,Double>  occurrenceProbability;
	private static ArrayList<Integer> failedTestCases;
	private static ArrayList<Integer> failedTestCases_OT;
	private static ArrayList<Integer> selectedTestCasesRT;
	private static PrintStream output;
	
	public static void main(String args[]) throws IOException, InterruptedException{

		int totalTests;  
		int numberOfPartitions;
	
		String baseDir = System.getProperty("user.dir");
		System.out.println("Base Directory: "+baseDir);
		String  program_name = args[0];
		String version_name = args[1];
		
		int[] domainSize;  	
		double[] p_i; 
		double significance;
		double desiredError;
		double learningFactor;
		
		switch (program_name) {
		case "grep":
			totalTests = 809;
			numberOfPartitions = 4;
			domainSize = new int[numberOfPartitions];
			p_i = new double[numberOfPartitions];
			domainSize[0] = 249; 
			domainSize[1] = 5;
			domainSize[2] = 181;  // FOR GREP: 809 tests
			domainSize[3] = 374;  // FOR GREP: 809 tests
			
			significance= 0.01;
			desiredError = 0.2;
			learningFactor=0.5;
			break;
		case "gzip":
			totalTests = 214; 
			numberOfPartitions = 5;
			domainSize = new int[numberOfPartitions];
			p_i = new double[numberOfPartitions];	
			domainSize[0] = 9; //GZIP - number of test cases assigned to each partition 
			domainSize[1] = 13;
			domainSize[2] = 83;   
			domainSize[3] = 83;   
			domainSize[4] = 26; 
		
			significance= 0.01;
			desiredError = 0.1;
			learningFactor=0.5;
			break;
		case "sed":
			totalTests = 363; 
			numberOfPartitions = 3;
			domainSize = new int[numberOfPartitions];
			p_i = new double[numberOfPartitions];	
			domainSize[0] = 7; //SED - number of test cases assigned to each partition 
			domainSize[1] = 19;
			domainSize[2] = 337;   
			
			significance= 0.01;
			desiredError = 0.2;
			learningFactor=0.5;
			break;
		case "flex":
			totalTests = 670; 
			numberOfPartitions = 6;
			domainSize = new int[numberOfPartitions];
			p_i = new double[numberOfPartitions];	
			domainSize[0] = 211; //FLEX - number of test cases assigned to each partition 
			domainSize[1] = 96;
			domainSize[2] = 96;   
			domainSize[3] = 96;   
			domainSize[4] = 96;   
			domainSize[5] = 75;   
			
			significance= 0.01;
			desiredError = 0.2;
			learningFactor=0.5;
			break;
		default:
			throw new IllegalArgumentException("Invalid program name provided as input: "+program_name);		
		}
			
		// CLEAN RESULTS DIRECTORY
		File dir = new File(baseDir+"/results");
		for (File file: dir.listFiles()){
			if(!file.isDirectory())
				file.delete();
		}
		
		String faultMatrix = null;
		String faultMatrixID = null;
		String coverageCriterion = null;
		
		int repetitions=50;
		int numberOf_RT_wins;
		int numberOf_OT_wins;
		int numberOf_ties;
		// START EXPERIMENTS
		PrintWriter pw = null;
		PrintWriter csv_w = null;
		PrintWriter csv_2_w = null;
		PrintWriter csv_3_w = null;
		PrintWriter csv_4_w = null;
		PrintWriter csv_5_w = null;
		StringBuilder sb;
		StringBuilder sb2;
		StringBuilder sb3;
		StringBuilder sb4;
		StringBuilder sb5;
		try{
		for (int expCombination=1;expCombination<7;expCombination++){
			System.out.println("START EXPERIMENT "+expCombination);
			sb = new StringBuilder();
			sb2 = new StringBuilder();
			sb3 = new StringBuilder();
			sb4 = new StringBuilder();
			sb5 = new StringBuilder();
			switch (expCombination){
			case 1:	
				faultMatrix = baseDir+"/subjects/"+program_name+"/"+version_name+"/fault_matrices/fault_matrix_hard_key_tc.txt";
				faultMatrixID = "hardFM";
				coverageCriterion = "function";
				break;
			case 2:
				faultMatrix = baseDir+"/subjects/"+program_name+"/"+version_name+"/fault_matrices/fault_matrix_hard_key_tc.txt";
				faultMatrixID = "hardFM";
				coverageCriterion = "branch";
				break;
			case 3:
				faultMatrix = baseDir+"/subjects/"+program_name+"/"+version_name+"/fault_matrices/fault_matrix_hard_key_tc.txt";
				faultMatrixID = "hardFM";
				coverageCriterion = "line";
				break;
			case 4:
				faultMatrix = baseDir+"/subjects/"+program_name+"/"+version_name+"/fault_matrices/fault_matrix_key_tc.txt";
				faultMatrixID = "defaultFM";
				coverageCriterion = "function";
				break;
			case 5:
				faultMatrix = baseDir+"/subjects/"+program_name+"/"+version_name+"/fault_matrices/fault_matrix_key_tc.txt";
				faultMatrixID = "defaultFM";
				coverageCriterion = "branch";
				break;
			case 6:
				faultMatrix = baseDir+"/subjects/"+program_name+"/"+version_name+"/fault_matrices/fault_matrix_key_tc.txt";
				faultMatrixID = "defaultFM";
				coverageCriterion = "line";
			
				break;
			}
		
			numberOf_RT_wins=0;
			numberOf_OT_wins=0;
			numberOf_ties=0;
			
			String outFile = baseDir+"/results/output_"+program_name+"_"+version_name+"_"+coverageCriterion+"_"+faultMatrixID+".txt";
			//CSV Files 1 and 2 are created within the Experiment constructor. CSV file 3 is constructed hereafter
			String csv_3 = outFile.replace("txt", "csv");
			csv_3 = csv_3.replace("output", "csv_Exp");
			System.out.println("csv_3 "+csv_3);
			
			Experiment ex = new Experiment(totalTests, numberOfPartitions, program_name, version_name, domainSize, outFile, significance, desiredError, learningFactor,coverageCriterion, faultMatrix);
			pw = ex.getPw();
			csv_w = ex.getCsv_w();
			csv_2_w = ex.getCsv_w_2();
			csv_3_w = new PrintWriter(new BufferedWriter(new FileWriter(csv_3)));
			csv_4_w  = ex.getCsv_w_4();
			csv_5_w  = ex.getCsv_w_5(); 
			sb3.append("repetition");
			sb3.append(",");
			sb3.append("RTwin");
			for(int k=0;k<numberOfPartitions;k++){
				sb3.append(",");
				sb3.append("profileP_"+k);
				}			
			sb3.append("\n");
			
			
			for(int i=0; i<repetitions ;i++){	
				p_i = generateProfile(numberOfPartitions);
				ex.setProfile(p_i);
				System.out.println("\n\n/***********************RUN NUMBER "+i+"************************/\n\n");
				pw.println("\n\n/***********************RUN NUMBER "+i+"************************/\n\n");
				//PRINT PROFILE FOR THIS RUN
				for (int j=0; j<numberOfPartitions;j++){
					pw.println("Operational Profile. Partition "+j+" : "+p_i[j]);
					System.out.println("Operational Profile. Partition "+j+" : "+p_i[j]);
				}
				pw.flush();
				int out = ex.runSession();
				if(out==1)
					numberOf_RT_wins++;
				if (out==0)
					numberOf_OT_wins++;
				if (out==-1)
					numberOf_ties++;
				if (out==-2){
					System.out.println("ERROR: the return value must be either 1, 0 or -1 ");;
					System.exit(0);
				}
					
					
				System.out.println("\n\n/***********************ENDED REPETITION "+i+"************************/\n\n");
				sb.append(ex.getCsv_string(0, i));
				sb2.append(ex.getCsv_string(1, i));
				sb4.append(ex.getCsv_string(2, i));
				sb5.append(ex.getCsv_string(3, i));
				
				
				sb3.append(i);
				sb3.append(",");
				sb3.append(out);
				for (int k=0;k<numberOfPartitions;k++){
					sb3.append(",");
					sb3.append(p_i[k]);
				}
				sb3.append("\n");
				csv_3_w.write(sb3.toString());
				csv_3_w.flush();
				sb3 = new StringBuilder();
				
				csv_w.write(sb.toString());
				csv_w.flush();
				csv_2_w.write(sb2.toString());
				csv_2_w.flush();
				csv_4_w.write(sb4.toString());
				csv_4_w.flush();
				csv_5_w.write(sb5.toString());
				csv_5_w.flush();
				
				sb = new StringBuilder();
				sb2 = new StringBuilder();
				sb4 = new StringBuilder();
				sb5 = new StringBuilder();
			}
			csv_w.close();
			csv_2_w.close();	
			csv_3_w.close();
			csv_4_w.close();
			csv_5_w.close();
			
			System.out.println("Number of Wins of Reliability Testing "+numberOf_RT_wins);
			System.out.println("Number of  Wins of Operational Testing "+(numberOf_OT_wins));
			System.out.println("Number of Ties "+(numberOf_ties));
			pw.println("Number of Wins of Reliability Testing "+numberOf_RT_wins);
			pw.println("Number of Wins of Operational Testing "+(numberOf_OT_wins));
			pw.println("Number of Ties "+(numberOf_ties));
			pw.flush();
		} // END the 6 combinations of EXPERIMENTS
		System.out.println("END OF ALL EXPERIMENTS");
		}
		catch(Exception e){
			System.out.println("Exception uncaught previously ");
			e.printStackTrace();
		}
		finally{
			pw.flush();
			pw.close();
			csv_w.flush();
			csv_w.close();
			csv_2_w.flush();
			csv_2_w.close();
			csv_3_w.flush();
			csv_3_w.close();
			csv_4_w.flush();
			csv_4_w.close();
			csv_5_w.flush();
			csv_5_w.close();
		
		}
		
	}

	private static double[] generateProfile(int numberOfPartitions) {
		
		double[] rand = new double[numberOfPartitions];
		double sum = 0;
		for (int i=0;i<numberOfPartitions;i++){
			 rand[i] = Math.random();
			 sum = sum + rand[i];
		}
		for (int i=0;i<numberOfPartitions;i++)
			 rand[i] = rand[i]/sum;
			 
	/*	double sum2=0;
		for (int i=0;i<numberOfPartitions;i++){
			System.out.println("rand[i] "+rand[i]);
			sum2 =  sum2 + rand[i];
		}
		System.out.println("sum "+sum2);*/
		
		return rand;
	}

		
}
