'''
Created on Oct 1, 2015

@author: breno
'''

import logging
from operator import itemgetter
from random import choice


def update_coverage_information(missing_coverage_aux, tc_cov_dict, selected_tc, mode):
    covered = set(dict(tc_cov_dict[selected_tc]))
    missing = set(missing_coverage_aux)
    diff = missing.difference(covered)
    if diff:
        return list(diff)
    else:
        return []
        
def greedy_additional_get_next_tc(tc_list, missing_coverage, tc_cov_dict, mode, entities_importance=None, accept_zero_cov_flag=False, weights=None):
    """
    tc_list: a pool of test cases that can be selected
    missing_coverage: a list containing the entities yet to be covered 
    tc_cov_dict: a dictionary containing coverage information for each of the test cases available in tc_list 
    """
    next_tc = None
    ranks = []
    if mode == 'traditional':
        for tc in tc_list:
            tc_rank = 0
            aux_dict = dict(tc_cov_dict[tc])
            #print 'aux_dict: ', aux_dict
            for ent_id in missing_coverage:
                if ent_id in aux_dict:
                    tc_rank += 1
            ranks.append((tc, tc_rank))

    elif mode == 'weighted':
        assert weights != None, "'Weights' dict needs to be provided!"  
        for tc in tc_list:
            total_contrib = 0.0
            aux_set = set(dict(tc_cov_dict[tc]))
            ent_h = set(missing_coverage).intersection(entities_importance['high'])
            ent_m = set(missing_coverage).intersection(entities_importance['medium'])
            ent_l = set(missing_coverage).intersection(entities_importance['low'])
            c_h = len(aux_set.intersection(ent_h)) * weights['high']
            c_m = len(aux_set.intersection(ent_m)) * weights['medium']
            c_l = len(aux_set.intersection(ent_l)) * weights['low']
            total_contrib = c_h + c_m + c_l   
            tc_rank = round(total_contrib, 3)
            ranks.append((tc, tc_rank))

    aux = sorted(ranks, key=itemgetter(1), reverse=True)
    #===========================================================================
    max_cov = aux[0][1]
    candidates = [tc_tuple for tc_tuple in aux if tc_tuple[1] == max_cov]
    #if candidates > 1:
        #print "candidates: ", [tc[0] for tc in candidates]
    selected_tc_tuple = choice(candidates)  # In case of a tie, pick a random test case
    #selected_tc_tuple = aux[0]  # In case of a tie, pick the first test case (ordered by ID)
    #===========================================================================    
    selected_tc_id = selected_tc_tuple[0]
    selected_tc_cov = selected_tc_tuple[1]
    if selected_tc_cov > 0:
        next_tc = selected_tc_id
    elif selected_tc_cov == 0 and accept_zero_cov_flag:
        next_tc = selected_tc_id
    return next_tc

def select_greedy_additional(tc_list, missing_coverage, tc_cov_dict, tcs_that_trigger_mff, mode, entities_importance=None):
    '''If the most frequent fault is triggered, the selection procedure stops and the test suite is returned'''
    tc_list_aux = tc_list[:]
    if type(missing_coverage) == list:
        missing_coverage_aux = missing_coverage[:]
    if type(missing_coverage) == dict:
        missing_coverage_aux = missing_coverage.copy()
    ts = []
    while tc_list_aux:
        next_tc = greedy_additional_get_next_tc(tc_list_aux, missing_coverage_aux, tc_cov_dict, mode, entities_importance)
        if next_tc:
            ts.append(next_tc)
            tc_list_aux.remove(next_tc)
            missing_coverage_aux = update_coverage_information(missing_coverage_aux, tc_cov_dict, next_tc, mode)        
            if next_tc in tcs_that_trigger_mff:
                return ts        
            elif next_tc not in tcs_that_trigger_mff and (missing_coverage_aux == [] or missing_coverage_aux == {}):
                logging.debug("full {} coverage has been achieved for OP before hitting the target fault".format(mode))
                return ts
        else:
            logging.debug("full {} coverage for the target entities has been achieved".format(mode))
            return ts

def select_n_greedy_additional(numtcs, tc_list, missing_coverage, tc_cov_dict, mode, entities_importance=None, weights=None):
    '''Selects n test cases from tc_list'''        
    def _reset_missing_coverage():
        if type(missing_coverage) == list:
            missing_coverage_aux = missing_coverage[:]
        if type(missing_coverage) == dict:
            missing_coverage_aux = missing_coverage.copy()
        #logging.info("Missing coverage has been reseted...")
        return missing_coverage_aux
    
    if mode == 'weighted':
        assert weights != None, "'Weights' dict needs to be provided!"
    
    tc_list_aux = tc_list[:]
    missing_coverage_aux = None
    ts = []
    #===========================================================================
    # As per Roberto's request, if the number of tcs to be selected is bigger
    # than the number of tcs available in the pool, all the test cases currently
    # available in the pool are returned (prioritized) instead of raising an 
    # exception. The next line has been commented.
    #===========================================================================
    #assert numtcs <= len(tc_list_aux), "Error! Should select {} test cases but the pool contains {}".format(numtcs, len(tc_list_aux))
    if numtcs > len(tc_list_aux):
        logging.info("ATTENTION: want to select {} test cases but the pool contains {}".format(numtcs, len(tc_list_aux)))
        numtcs = len(tc_list_aux)
        logging.info("new numtcs: {}".format(numtcs))        
    #===========================================================================
    accept_zero_cov_flag = False
    count_next_tc_none = 0
    while len(ts) < numtcs:
        if count_next_tc_none >= 2:
            accept_zero_cov_flag = True
            logging.info("Will accept test cases that cover 0 relevant entities...")
        logging.debug("Trying to select tc #{}".format(len(ts)+1))
        if not missing_coverage_aux:
            missing_coverage_aux = _reset_missing_coverage()
        next_tc = greedy_additional_get_next_tc(tc_list_aux, missing_coverage_aux, tc_cov_dict, mode, entities_importance, accept_zero_cov_flag, weights)
        if next_tc:
            ts.append(next_tc)
            tc_list_aux.remove(next_tc)
            missing_coverage_aux = update_coverage_information(missing_coverage_aux, tc_cov_dict, next_tc, mode)
            if not missing_coverage_aux:
                logging.debug("Full coverage has been achieved after {}. Reseting coverage...".format(len(ts)))
            count_next_tc_none = 0
        else:
            count_next_tc_none += 1       
            if missing_coverage_aux:
                logging.debug("No test cases could cover the remaining missing entities (full coverage cannot be achieved considering this pool of test cases). Need to reset coverage...")
                missing_coverage_aux = _reset_missing_coverage()
    assert len(ts) == numtcs, "Something went wrong! Could not select {} test cases".format(numtcs)            
    return ts
        

if __name__ == '__main__':
    pass